import click
import yaml
import json
import traceback

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from . import manager


@click.group()
def instance():
    """Manage CKAN Instances"""
    pass


@instance.command()
@click.argument('INSTANCE_TYPE')
@click.argument('VALUES_FILE')
@click.option('--instance-id')
@click.option('--instance-name')
@click.option('--exists-ok', is_flag=True)
@click.option('--dry-run', is_flag=True)
@click.option('--update', 'update_', is_flag=True)
@click.option('--wait-ready', is_flag=True)
@click.option('--skip-deployment', is_flag=True)
@click.option('--skip-route', is_flag=True)
@click.option('--force', is_flag=True)
def create(instance_type, values_file, instance_id, instance_name, exists_ok, dry_run, update_, wait_ready,
           skip_deployment, skip_route, force):
    '''
    Create CKAN instance
    '''
    manager.create(instance_id=instance_id, instance_type=instance_type, instance_name=instance_name,
                   values_filename=values_file, exists_ok=exists_ok, dry_run=dry_run, update_=update_,
                   wait_ready=wait_ready, skip_deployment=skip_deployment, skip_route=skip_route, force=force)

    logs.exit_great_success()


@instance.command()
@click.argument('INSTANCE_ID_OR_NAME')
@click.argument('OVERRIDE_SPEC_JSON', required=False)
@click.option('--persist-overrides', is_flag=True)
@click.option('--wait-ready', is_flag=True)
@click.option('--skip-deployment', is_flag=True)
@click.option('--skip-route', is_flag=True)
@click.option('--force', is_flag=True)
def update(instance_id_or_name, override_spec_json, persist_overrides, wait_ready, skip_deployment, skip_route, force):
    """Update an instance to the latest resource spec, optionally applying the given json override to the resource spec

    Examples:

    ckan-cloud-operator ckan instance update <INSTANCE_ID_OR_NAME> '{"siteUrl": "http://localhost:5000"}' --wait-ready

    ckan-cloud-operator ckan instance update <INSTANCE_ID_OR_NAME> '{"replicas": 3}' --persist-overrides
    """
    override_spec = json.loads(override_spec_json) if override_spec_json else None
    manager.update(instance_id_or_name, override_spec=override_spec, persist_overrides=persist_overrides,
                   wait_ready=wait_ready, skip_deployment=skip_deployment, skip_route=skip_route,
                   force=force)
    logs.exit_great_success()


@instance.command()
@click.argument('INSTANCE_ID_OR_NAME')
@click.argument('ATTR', required=False)
@click.option('--with-spec', is_flag=True)
def get(instance_id_or_name, attr, with_spec):
    """Get detailed information about an instance, optionally returning only a single get attribute

    Example: ckan-cloud-operator ckan instance get <INSTANCE_ID_OR_NAME> deployment
    """
    if attr == 'spec':
        with_spec = True
    logs.print_yaml_dump(manager.get(instance_id_or_name, attr, with_spec=with_spec), exit_success=True)


@instance.command()
@click.argument('INSTANCE_ID_OR_NAME')
def edit(instance_id_or_name):
    '''Edit CKAN instance
    '''
    manager.edit(instance_id_or_name)


@instance.command('list')
@click.option('-f', '--full', is_flag=True)
@click.option('-q', '--quick', is_flag=True)
@click.option('-c', '--credentials', is_flag=True)
@click.option('--name')
def list_instances(full, quick, name, credentials):
    '''
    List existing CKAN instaces
    '''
    instances = list(
        manager.list_instances(full=full, quick=quick,
                               name=name, withCredentials=credentials)
    )
    logs.print_yaml_dump(instances)
    logs.exit_great_success(quiet=True)


# @instance.command()
# @click.argument('INSTANCE_ID')
# @click.argument('INSTANCE_NAME')
# def set_name(instance_id, instance_name):
#     logs.info(f'{instance_name} --> {instance_id}')
#     manager.set_name(instance_id, instance_name)
#     logs.exit_great_success()


@instance.command()
@click.argument('INSTANCE_ID_OR_NAME', nargs=-1)
@click.option('--no-dry-run', is_flag=True)
def delete(instance_id_or_name, no_dry_run):
    '''
    Delete CKAN instance
    '''
    generator = manager.delete_instances(instance_ids_or_names=instance_id_or_name, dry_run=not no_dry_run)
    while True:
        try:
            logs.info('Deleting instance', **next(generator))
        except StopIteration:
            break
        logs.info(**next(generator))
    logs.exit_great_success(quiet=True)


# @instance.command()
# @click.argument('INSTANCE_NAME')
# def delete_name(instance_name):
#     manager.delete_name(instance_name=instance_name)
#     logs.exit_great_success()
#
#
## Moved to `sysadmin`
# @instance.command()
# @click.argument('INSTANCE_ID_OR_NAME')
# @click.argument('NAME')
# @click.argument('EMAIL', required=False)
# @click.argument('PASSWORD', required=False)
# @click.option('--dry-run', is_flag=True)
# def create_ckan_admin_user(instance_id_or_name, name, email, password, dry_run):
#     logs.print_yaml_dump(manager.create_ckan_admin_user(instance_id_or_name, name, email, password, dry_run))
#     logs.exit_great_success()



@instance.command('logs')
@click.argument('INSTANCE_ID')
@click.option('--service', help='Service name. One of `ckan`, `giftless`, `jobs`, `jobs-db`, `redis`. Defaults to `ckan`', default='ckan')
@click.option('--since', help='Only return logs newer than a relative duration like 5s, 2m, or 3h. Defaults to all logs.')
@click.option('--follow', help='Specify if the logs should be streamed.')
@click.option('--tail', help='Lines of recent log file to display. Defaults to -1 with no selector, showing all log lines otherwise 10, if a selector is provided.')
@click.option('--container', help='Conainer name if multiple')
def ckan_logs(instance_id, **kubectl_args):
    '''
    Check CKAN and other service container logs
    '''
    manager.get_container_logs(instance_id, **kubectl_args)

@instance.command('ckan-exec')
@click.argument('INSTANCE_ID')
@click.option('--command', help='command to pass down to ckan CLI, without path to config file')
@click.option('--use-paster', help='Use paster over ckan CLI (supported in ckan v2.9)', default=False)
def ckan_exec(instance_id, command, use_paster):
    '''
    Executes ckan CLI commands

    See full list of command in [CKAN Docs](https://docs.ckan.org/en/2.9/maintaining/cli.html#ckan-commands-reference). You will need to pass full command as a string

    \b
    cco ckan instance ckan-exec --command='view clean --yes'
    cco ckan instance ckan-exec --command='jobs list'
    cco ckan instance ckan-exec --command='dataset show dataset-id'
    '''
    manager.run_ckan_commands(instance_id, command)


@instance.command('ssh')
@click.argument('INSTANCE_ID')
@click.option('--service', help='Service name. One of `ckan`, `giftless`, `jobs`, `jobs-db`, `redis`. Defaults to `ckan`', default='ckan')
@click.option('--command', help='One of `bash`, `sh`. Defaults to `bash`', default='bash')
def ckan_ssh(instance_id, service, command):
    '''
    SSH into the running container.
    '''
    manager.ssh_into_container(instance_id, service, command)


@instance.command('shell')
@click.option('--service', help='Service name. One of `ckan`, `giftless`, `jobs`, `jobs-db`, `redis`. Defaults to `ckan`')
@click.option('--command', help='One of `bash`, `sh`. Defaults to `bash`')
def ckan_shell(service, command):
    '''
    Run Unix-like operating system commands into the running container

    cco ckan instance shell --command='cat /etc/ckan/production.ini'
    '''
    pass


@click.group()
def sysadmin():
    """Create or delete system administrator for CKAN instance"""
    pass

instance.add_command(sysadmin)


@sysadmin.command('add')
@click.argument('INSTANCE_ID')
@click.option('--username', required=True, help='User name for user if user does not exist')
@click.option('--password', help='Password for user if user does not exist')
@click.option('--email', help='Valid Email address for user if user does not exist')
@click.option('--use-paster', help='Use paster over ckan CLI (supported in ckan v2.9)', default=False)
def sysadmin_add(instance_id, username, password, email, use_paster):
    '''
    Creates or makes given user system administrator

    cco ckan instance sysadmin add USERNAME --pasword pasword --email email@email.com
    '''
    manager.create_ckan_admin_user(instance_id, username, password, email, use_paster)


@sysadmin.command('rm')
@click.argument('INSTANCE_ID')
@click.option('--username', required=True, help='Passowrd for user if user does not exist')
@click.option('--use-paster', help='Use paster over ckan CLI (supported in ckan v2.9)', default=False)
def sysadmin_rm(instance_id, username, use_paster):
    '''
    Removes System administrator privileges from given user

    cco ckan instance sysadmin rm USERNAME
    '''
    manager.delete_ckan_admin_user(instance_id, username, use_paster)


@click.group()
def solr():
    """Update, clear or check search index for CKAN instance"""
    pass

instance.add_command(solr)


@solr.command('check')
@click.argument('INSTANCE_ID')
def solr_check(instance_id):
    '''
    Check the search index
    '''
    manager.run_solr_commands(instance_id, 'check')

@solr.command('clear')
@click.argument('INSTANCE_ID')
def solr_clear(instance_id):
    '''
    Clear the search index
    '''
    manager.run_solr_commands(instance_id, 'clear')

@solr.command('rebuild')
@click.argument('INSTANCE_ID')
def solr_rebuild(instance_id):
    '''
    Rebuild search index
    '''
    manager.run_solr_commands(instance_id, 'rebuild')

@solr.command('rebuild-fast')
@click.argument('INSTANCE_ID')
def solr_rebuild_fast(instance_id):
    '''
    Reindex with multiprocessing
    '''
    manager.run_solr_commands(instance_id, 'check')

@solr.command('show')
@click.argument('INSTANCE_ID')
@click.option('--dataset-id', help='Dataset name to show index for')
def solr_show(instance_id, dataset_id):
    '''
    show --dataset-id=dataset-id-or-name
    '''
    manager.run_solr_commands(instance_id, 'show', dataset_id=dataset_id)

@click.group()
def deployment():
    """Create, Deploy and manage CKAN instances on Kubernetes Cluster"""
    pass

instance.add_command(deployment)

@deployment.command('status')
def deployment_status():
    '''
    Shows status of the deployment. Result of `helm status release-name`
    '''


@deployment.command('logs')
def deployment_logs():
    '''
    Shows deployment logs
    '''

@deployment.command('version')
def deployment_version():
    '''
    Shows version of the latest successful deployment. Same as https://site-url/version
    '''


@click.group('image')
def deployment_image():
    '''
    Create, Deploy and manage CKAN instances on Kubernetes Cluster
    '''
    pass

deployment.add_command(deployment_image)

@deployment_image.command('get')
@click.option('--service', help='Service name. One of `ckan`, `giftless`, `jobs`, `jobs-db`, `redis`. Defaults to `ckan`')
def deployment_image_get(service):
    '''
    Get and set CKAN or Related service images.
    '''

@deployment_image.command('set')
@click.argument('IMAGE_NAME')
@click.option('--service', help='Service name. One of `ckan`, `giftless`, `jobs`, `jobs-db`, `redis`. Defaults to `ckan`')
def deployment_image_set(image_name, service):
    '''
    Force set given image for the given service
    '''


@click.group()
def infra():
    '''
    Manage and debug CKAN related infrastructure like SOLR Cloud and Postgres Databases
    '''
    pass

instance.add_command(infra)


@click.group('solr')
def deployment_solr():
    '''
    Check logs of SolrCloud service and restart them.
    '''
    pass

infra.add_command(deployment_solr)


@deployment_solr.command('logs')
@click.option('--zookeper-only', help='Make operations only for zookeper pods', is_flag=True)
@click.option('--solrcloud-only', help='Make operations only for solrcloud pods', is_flag=True)
@click.option('--since', help='Only return logs newer than a relative duration like 5s, 2m, or 3h. Defaults to all logs.')
@click.option('--follow', help='Specify if the logs should be streamed.')
@click.option('--tail', help='Lines of recent log file to display. Defaults to -1 with no selector, showing all log lines otherwise 10, if a selector is provided.')
@click.option('--container', help='Conainer name if multiple')
@click.option('--grep', help='Filter logs by the given word (case insensitive)')
def infra_solr_logs(zookeper_only, solrcloud_only, since, follow, tail, container, grep):
    '''
    See logs of SolrCloud and ZooKeeper containers
    '''
    pass


@deployment_solr.command('restart')
@click.option('--zookeper-only', help='Make operations only for zookeper pods', is_flag=True)
@click.option('--solrcloud-only', help='Make operations only for solrcloud pods', is_flag=True)
def infra_solr_restart(zookeper_only, solrcloud_only):
    '''
    Restart SolrCloud and Zookeeper containers
    '''
    pass


@click.group('db')
def deployment_db():
    '''
    Get database connection string
    '''
    pass

infra.add_command(deployment_db)


@deployment_db.command('get')
def infra_solr_restart():
    '''
    Get master connection string for ckan Database
    '''
    pass



import unittest
from unittest.mock import patch, MagicMock

from ckan_cloud_operator.routers.annotations import CkanRoutersAnnotations
from ckan_cloud_operator.routers import manager


class RoutersManagerTestCase(unittest.TestCase):
    @patch('ckan_cloud_operator.routers.manager.get')
    @patch('ckan_cloud_operator.kubectl.get_resource')
    @patch('ckan_cloud_operator.routers.traefik.manager.create')
    @patch('ckan_cloud_operator.providers.cluster.manager.get_or_create_multi_user_volume_claim')
    @patch.object(CkanRoutersAnnotations, 'json_annotate')
    def test_create_with_traefik_router(self, json_annotate, muvc, create, get_resource, get):
        spec = {
            'type': 'traefik',
            'default-root-domain': 'test.local'
        }
        manager.create('datapusher', spec)

        get.assert_called_once_with('datapusher', only_dns=True, failfast=True)
        get_resource.assert_called_once_with(
            'stable.viderum.com/v1',
            'CkanCloudRouter',
            'datapusher',
            {'ckan-cloud/router-name': 'datapusher', 'ckan-cloud/router-type': 'traefik'},
            spec=spec
        )
        muvc.assert_called_once_with({'router-name': 'datapusher', 'router-type': 'traefik'})
        json_annotate.assert_called_once_with('default-root-domain', 'test.local')

    def test_create_with_unsupported_router(self):
        spec = {
            'type': 'wildtraefik',
            'default-root-domain': 'test.local'
        }
        with self.assertRaisesRegex(AssertionError, 'Invalid router spec'):
            manager.create('datapusher', spec)

    def test_create_without_default_root_domain(self):
        spec = {'type': 'traefik'}
        with self.assertRaisesRegex(AssertionError, 'Invalid router spec'):
            manager.create('datapusher', spec)

    def test_get_traefik_router_spec(self):
        expected_spec = {
            'type': 'traefik',
            'default-root-domain': 'default',
            'cloudflare': {
                'email': 'default',
                'api-key': 'default'
            },
            'wildcard-ssl-domain': None,
            'external-domains': False,
            'dns-provider': None
        }
        self.assertEqual(manager.get_traefik_router_spec(), expected_spec)

        expected_spec = {
            'type': 'traefik',
            'default-root-domain': 'test.local',
            'cloudflare': {
                'email': 'admin@test.local',
                'api-key': 'my-super-secret-api-key'
            },
            'wildcard-ssl-domain': 'test.local',
            'external-domains': True,
            'dns-provider': 'CloudFlare'
        }
        options = {
            'default_root_domain': 'test.local',
            'cloudflare_email': 'admin@test.local',
            'cloudflare_api_key': 'my-super-secret-api-key',
            'wildcard_ssl_domain': 'test.local',
            'external_domains': True,
            'dns_provider': 'CloudFlare'
        }
        self.assertEqual(manager.get_traefik_router_spec(**options), expected_spec)

    @patch('ckan_cloud_operator.routers.manager._init_router')
    @patch('ckan_cloud_operator.routers.traefik.manager')
    @patch('ckan_cloud_operator.routers.manager.routes_manager.list')
    def test_update(self, list, traefik_manager, _init_router):
        _init_router.return_value = 'router', {'update': True}, 'traefik', {}, {}, {'manager': traefik_manager}
        list.return_value = 'router'
        manager.update('datapusher', wait_ready=True)
        _init_router.assert_called_once_with('datapusher')
        list.assert_called_once_with({})
        traefik_manager.update.assert_called_once_with('datapusher', True, {'update': True}, {}, 'router', dry_run=False)

    @patch('ckan_cloud_operator.kubectl.get')
    def test_list(self, get):
        get.return_value = {
            'items': [
                {
                    'metadata': {
                        'name': 'datapusher',
                    },
                    'spec': {
                        'type': 'traefik'
                    }
                },
                {
                    'metadata': {
                        'name': 'infra-1',
                    },
                    'spec': {
                        'type': 'traefik'
                    }
                },
                {
                    'metadata': {
                        'name': 'prod',
                    },
                    'spec': {
                        'type': 'traefik'
                    }
                }
            ]
        }
        expected_list = [
            {'name': 'datapusher', 'type': 'traefik'},
            {'name': 'infra-1', 'type': 'traefik'},
            {'name': 'prod', 'type': 'traefik'}
        ]
        self.assertEqual(manager.list(values_only=True, async_print=False), expected_list)

    @patch('ckan_cloud_operator.kubectl.get')
    @patch('ckan_cloud_operator.routers.manager._init_router')
    @patch('ckan_cloud_operator.routers.traefik.manager')
    @patch('ckan_cloud_operator.routers.manager.routes_manager.list')
    def test_get(self, list, traefik_manager, _init_router, get):
        get.return_value = {'metadata': {'annotations': 'annotations'}, 'spec': {'type': 'test'}}
        _init_router.return_value = 'router', {'update': True}, 'traefik', {}, {}, {'manager': traefik_manager}
        list.return_value = [{'spec': {'type': 'CkanCloudRouter'}}]
        traefik_manager.get.return_value = {'type': 'CkanCloudRouter'}

        expected_result = {
            'name': 'datapusher',
            'annotations': 'annotations',
            'routes': [{'type': 'CkanCloudRouter'}],
            'type': 'traefik',
            'deployment': {'type': 'CkanCloudRouter'},
            'ready': False,
            'dns': {'type': 'CkanCloudRouter'},
            'spec': {'ready': True, 'type': 'test'}
        }
        self.assertEqual(manager.get('datapusher'), expected_result)

        expected_result = {
            'name': 'datapusher',
            'dns': {'type': 'CkanCloudRouter'}
        }
        self.assertEqual(manager.get('datapusher', only_dns=True), expected_result)

    @patch('ckan_cloud_operator.routers.manager.get_env_id')
    @patch('ckan_cloud_operator.routers.manager._init_router')
    @patch('ckan_cloud_operator.kubectl.get_resource')
    @patch('ckan_cloud_operator.kubectl.apply')
    @patch('ckan_cloud_operator.routers.traefik.manager')
    @patch('ckan_cloud_operator.routers.manager.hashlib')
    def test_create_subdomain_route(self, hashlib, traefik_manager, apply, get_resource, _init_router, get_env_id):
        get_env_id.return_value = 'p'
        _init_router.return_value = 'router', {'update': True}, 'traefik', {}, {}, {'manager': traefik_manager}
        hashmock = MagicMock()
        hashmock.hexdigest = lambda: '123123123hashed'
        hashlib.sha3_224.return_value = hashmock
        route_spec = {
            'target-type': 'datapusher',
            'sub-domain': 'www',
            'root-domain': 'example.com',
            'datapusher-name': 'datapusher-1'
        }
        manager.create_subdomain_route('datapusher', route_spec)

        spec = {
            'name': 'cc123123123hashed',
            'type': 'datapusher-subdomain',
            'root-domain': 'example.com',
            'sub-domain': 'www',
            'router_name': 'datapusher',
            'router_type': 'traefik',
            'route-target-type': 'datapusher',
            'route-target-resource-id': 'datapusher-1',
            'datapusher-name': 'datapusher-1'
        }
        labels = {
            'ckan-cloud/route-type': 'datapusher-subdomain',
            'ckan-cloud/route-root-domain': 'example.com',
            'ckan-cloud/route-sub-domain': 'www',
            'ckan-cloud/route-target-type': 'datapusher',
            'ckan-cloud/route-target-resource-id': 'datapusher-1',
            'ckan-cloud/route-datapusher-name': 'datapusher-1'
        }
        get_resource.assert_called_once_with('stable.viderum.com/v1', 'CkanCloudRoute', 'cc123123123hashed', labels, spec=spec)

    def test_create_subdomain_route_with_invalid_spec(self):
        route_spec = {
            'target-type': 'infra-1'
        }
        with self.assertRaisesRegex(Exception, 'Invalid route spec'):
            manager.create_subdomain_route('datapusher', route_spec)

    @patch('ckan_cloud_operator.kubectl.install_crd')
    @patch('ckan_cloud_operator.routers.manager.routes_manager.install_crds')
    def test_install_crds(self, install_crds, install_crd):
        manager.install_crds()
        install_crd.assert_called_once_with('ckancloudrouters', 'ckancloudrouter', 'CkanCloudRouter')
        install_crds.assert_called_once()

    @patch('ckan_cloud_operator.routers.manager._init_router')
    @patch('ckan_cloud_operator.routers.traefik.manager')
    def test_delete(self, traefik_manager, _init_router):
        _init_router.return_value = 'router', {'update': True}, 'traefik', {}, {}, {'manager': traefik_manager}
        manager.delete('datapusher')
        traefik_manager.delete.assert_called_once_with('datapusher')



import tempfile
import binascii
import os
import datetime

from ckan_cloud_operator import gcloud
from ckan_cloud_operator.infra import CkanInfra
from ckan_cloud_operator import kubectl
from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator import logs


PERMISSIONS_FUNCTION_PACKAGE_JSON = """
{
  "dependencies": {
    "@google-cloud/storage": "^2.3.4"
  }
}
"""


PERMISSIONS_FUNCTION_JS = lambda function_name, project_id, bucket_name: """
exports.""" + function_name + """ = (event, callback) => {
  const {Storage} = require('@google-cloud/storage');
  const projectId = '""" + project_id + """';
  const bucketName = '"""+ bucket_name +"""';
  const file_name = event.data.name;
  if (file_name === 'ckan') {
      callback();
  } else {
      console.log(`  File: ${file_name}`);
      const storage = new Storage();
      const bucket = storage.bucket(bucketName);
      const file = bucket.file(file_name)
      file.makePrivate().then(function(){
        callback();
      }).catch(function(err){
        callback(err);
      });
  };
};
"""


def deploy_storage_permissions_function():
    """Deploys a serverless function that sets all keys to private"""
    ckan_infra = CkanInfra()
    bucket_name = ckan_infra.GCLOUD_STORAGE_BUCKET
    project_id = ckan_infra.GCLOUD_AUTH_PROJECT
    function_name = bucket_name.replace('-', '') + 'permissions'
    function_js = PERMISSIONS_FUNCTION_JS(function_name, project_id, bucket_name)
    package_json = PERMISSIONS_FUNCTION_PACKAGE_JSON
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(f'{tmpdir}/package.json', 'w') as f:
            f.write(package_json)
        with open(f'{tmpdir}/index.js', 'w') as f:
            f.write(function_js)
        gcloud.check_call(
            f'functions deploy {function_name} '
                               f'--runtime nodejs6 '
                               f'--trigger-resource {bucket_name} '
                               f'--trigger-event google.storage.object.finalize '
                               f'--source {tmpdir} '
                               f'--retry '
                               f'--timeout 30s ',
            ckan_infra=ckan_infra
        )



def deploy_gcs_minio_proxy(router_name):
    """Deploys a minio proxy (AKA gateway) for access to google storage"""
    labels = {'app': 'ckan-cloud-gcsminio-proxy'}
    if not kubectl.get('secret gcsminio-proxy-credentials', required=False):
        print('Creating minio credentials')
        minio_access_key = binascii.hexlify(os.urandom(8)).decode()
        minio_secret_key = binascii.hexlify(os.urandom(12)).decode()
        kubectl.update_secret('gcsminio-proxy-credentials', {
            'MINIO_ACCESS_KEY': minio_access_key,
            'MINIO_SECRET_KEY': minio_secret_key,
        })
    kubectl.apply(kubectl.get_deployment('gcsminio-proxy', labels, {
        'replicas': 1,
        'revisionHistoryLimit': 10,
        'strategy': {'type': 'RollingUpdate',},
        'template': {
            'metadata': {
                'labels': labels,
                'annotations': {
                    'ckan-cloud/operator-timestamp': str(datetime.datetime.now())
                }
            },
            'spec': {
                'containers': [
                    {
                        'name': 'minio',
                        'image': 'viderum/ckan-cloud-operator-gcsminio-proxy',
                        'env': [
                            {
                                'name': 'GOOGLE_APPLICATION_CREDENTIALS',
                                'value': '/gcloud-credentials/credentials.json'
                            }
                        ],
                        'envFrom': [{'secretRef': {'name': 'gcsminio-proxy-credentials'}}],
                        'ports': [{'containerPort': 9000}],
                        'volumeMounts': [
                            {'name': 'gcloud-credentials',
                             'mountPath': '/gcloud-credentials/credentials.json',
                             'subPath': 'GCLOUD_SERVICE_ACCOUNT_JSON'},
                        ],
                    }
                ],
                'volumes': [
                    {'name': 'gcloud-credentials', 'secret': {'secretName': 'ckan-infra'}},
                ]
            }
        }
    }))
    service = kubectl.get_resource('v1', 'Service','gcsminio-proxy', labels)
    service['spec'] = {
        'ports': [
            {'name': '9000', 'port': 9000}
        ],
        'selector': labels
    }
    kubectl.apply(service)
    if not routers_manager.get_backend_url_routes('gcs-minio'):
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'backend-url',
            'target-resource-id': 'gcs-minio',
            'backend-url': 'http://gcsminio-proxy.ckan-cloud:9000',
            'sub-domain': 'default',
            'root-domain': 'default',
        })
    routers_manager.update(router_name, wait_ready=True)



"""Manage annotations associated with Kubernetes resources

Different types of annotations can be attached to any Kubernetes resource or custom resource

The annotations are stored in various places - secrets, labels, annotations and configmaps
"""


from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers import manager as providers_manager
from ckan_cloud_operator import logs

from .minio.constants import PROVIDER_ID as minio_provider_id
from .s3.constants import PROVIDER_ID as s3_provider_id
from .constants import PROVIDER_SUBMODULE, CONFIG_NAME


def initialize(interactive=False, provider_id=None, storage_suffix=None, use_existing_disk_name=None, dry_run=False):
    if not provider_id and interactive:
        config_manager.interactive_set(
            {'use-cloud-native-storage': True},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )
        if config_manager.get('use-cloud-native-storage', secret_name=CONFIG_NAME):
            provider_id = get_provider_id()

    logs.info(f'Storage init: chosen provider_id: {provider_id}')
    provider = get_provider(
        default=minio_provider_id,
        provider_id=provider_id
    ).initialize(
        interactive=interactive,
        storage_suffix=storage_suffix,
        use_existing_disk_name=use_existing_disk_name,
        dry_run=dry_run
    )
    if provider:
        provider.initialize(
            interactive=interactive,
            storage_suffix=storage_suffix,
            use_existing_disk_name=use_existing_disk_name,
            dry_run=dry_run
        )


def get_provider_id():
    from ckan_cloud_operator.providers.cluster import manager as cluster_manager
    from ckan_cloud_operator.providers.cluster.gcloud.constants import PROVIDER_ID as gcloud_provider_id
    from ckan_cloud_operator.providers.cluster.aws.constants import PROVIDER_ID as aws_provider_id
    from ckan_cloud_operator.providers.cluster.azure.constants import PROVIDER_ID as azure_provider_id
    from .s3.constants import PROVIDER_ID as s3_provider_id
    from .gcloud.constants import PROVIDER_ID as gcloud_storage_provider_id
    from .azure.constants import PROVIDER_ID as azure_storage_provider_id

    cluster_provider_id = cluster_manager.get_provider_id()
    if cluster_provider_id == gcloud_provider_id:
        return gcloud_storage_provider_id
    elif cluster_provider_id == aws_provider_id:
        return s3_provider_id
    elif cluster_provider_id == azure_provider_id:
        return azure_storage_provider_id
    return 'minio'


def get_provider(default=None, provider_id=None):
    return providers_manager.get_provider(PROVIDER_SUBMODULE, default=default, provider_id=provider_id, required=False)


def print_credentials(raw=False, storage_suffix=None, provider_id=None):
    return get_provider(provider_id=provider_id).print_credentials(raw=raw, storage_suffix=storage_suffix)



import click

from ckan_cloud_operator import logs

from . import manager


drone_manager = manager.Drone()

@click.group()
def drone():
    """Manage Drone CI/CD"""
    pass


@drone.command()
@click.option('--force-update', default=False, help='Force update drone configurations [default: develop]', is_flag=True)
def initialize(force_update):
    """Initialize drone"""
    drone_manager.initialize(force_update)



import click

from . import driver
from ckan_cloud_operator import logs


@click.group()
def helm():
    """Interact with Helm/Tiller"""
    pass


@helm.command()
@click.argument('TILLER_NAMESPACE_NAME')
def init(tiller_namespace_name):
    driver.init(tiller_namespace_name)
    logs.exit_great_success(quiet=True)



from ckan_cloud_operator.labels import manager as labels_manager


def update(instance_id, instance, dry_run=False):
    return _get_deployment_provider(instance).update(instance_id, instance, dry_run=dry_run)


def delete(instance_id, instance):
    return _get_deployment_provider(instance).delete(instance_id, instance)


def get(instance_id, instance):
    return _get_deployment_provider(instance).get(instance_id, instance)


def get_backend_url(instance_id, instance):
    deployment_provider = _get_deployment_provider(instance, required=False)
    if deployment_provider:
        return deployment_provider.get_backend_url(instance_id, instance)
    else:
        return None


def pre_update_hook(instance_id, instance, override_spec, skip_route, dry_run=False):
    return _get_deployment_provider(instance).pre_update_hook(instance_id, instance, override_spec, skip_route,
                                                              dry_run=dry_run)


def create_ckan_admin_user(instance_id, instance, user):
    _get_deployment_provider(instance).create_ckan_admin_user(instance_id, instance, user)


def _get_deployment_provider(instance, required=True):
    for get_provider in [_get_helm_deployment_provider]:
        res = get_provider(instance)
        if res:
            break
    if not res and required:
        raise Exception(f'failed to find provider for instance')
    else:
        return res


def _get_helm_deployment_provider(instance):
    label_prefix = labels_manager.get_label_prefix()
    if instance['metadata'].get('annotations', {}).get(f'{label_prefix}/deployment-provider') == 'helm':
        from .helm import manager as helm_manager
        return helm_manager
    else:
        return None



import subprocess
import yaml
import os
import json

from ckan_cloud_operator import kubectl
from ckan_cloud_operator.gitlab import CkanGitlab
from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.routers import manager as routers_provider
from ckan_cloud_operator.providers.ckan import manager as ckan_manager


def get_path_to_old_cluster_kubeconfig():
    return ckan_manager.get_path_to_old_cluster_kubeconfig()


def get_solr_config(old_site_id, path_to_old_cluster_kubeconfig):
    solr_collection_name = old_site_id.replace('-', 'dash')
    output = subprocess.check_output(f'KUBECONFIG={path_to_old_cluster_kubeconfig} '
                                     f'kubectl -n solr exec zk-0 zkCli.sh '
                                     f'get /collections/{solr_collection_name} 2>&1', shell=True)
    print(output)
    solr_config = None
    for line in output.decode().splitlines():
        if line.startswith('{"configName":'):
            solr_config = json.loads(line)["configName"]
    return solr_config


def get_instance_env(old_site_id, path_to_old_cluster_kubeconfig):
    output = subprocess.check_output(f'KUBECONFIG={path_to_old_cluster_kubeconfig} '
                                     f'kubectl -n {old_site_id} get deployment {old_site_id}-cmd -o yaml',
                                     shell=True)
    deployment = yaml.load(output)
    containers = deployment['spec']['template']['spec']['containers']
    assert len(containers) == 1, f'invalid number of containers {len(containers)}'
    container = containers[0]
    fetch_secret_name = None
    envvar_secrets = {}
    extra_values = {}
    for e in container['env']:
        name = e.pop('name')
        if 'valueFrom' in e:
            value_from = e.pop('valueFrom')
            assert len(e) == 0
            secret_key_ref = value_from.pop('secretKeyRef')
            assert len(value_from) == 0
            secret_key = secret_key_ref.pop('key')
            secret_name = secret_key_ref.pop('name')
            assert len(secret_key_ref) == 0
            assert not fetch_secret_name or fetch_secret_name == secret_name
            fetch_secret_name = secret_name
            envvar_secrets[name] = secret_key
        else:
            value = e.pop('value')
            assert len(e) == 0
            extra_values[name] = value
    output = subprocess.check_output(f'KUBECONFIG={path_to_old_cluster_kubeconfig} '
                                     f'kubectl -n {old_site_id} get secret {fetch_secret_name} -o yaml',
                                     shell=True)
    secret = kubectl.decode_secret(yaml.load(output))
    instance_env = {}
    for key, secret_key in envvar_secrets.items():
        instance_env[key] = secret[secret_key]
    return dict(instance_env, **extra_values)


def migrate_from_deis(old_site_id, new_instance_id, router_name, deis_instance_class,
                      skip_gitlab=False, db_migration_name=None, recreate=False,
                      skip_routes=False, skip_solr=False, skip_deployment=False,
                      no_db_proxy=False):
    assert db_migration_name, 'migration without a db migration is not supported yet'
    log_labels = {'instance': new_instance_id}
    if recreate:
        from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
        DeisCkanInstance(new_instance_id).delete(force=True, wait_deleted=not db_migration_name)
    logs.info(f'Migrating from old site id {old_site_id} to new instance id {new_instance_id}', **log_labels)
    instance_kind = ckan_manager.instance_kind()
    values = kubectl.get(f'{instance_kind} {new_instance_id}', required=False)
    if values:
        logs.info('instance already exists', **log_labels)
    else:
        logs.info('creating instance', **log_labels)
        path_to_old_cluster_kubeconfig = get_path_to_old_cluster_kubeconfig()
        solr_config = get_solr_config(old_site_id, path_to_old_cluster_kubeconfig)
        assert solr_config, 'failed to get solr config name'
        instance_env = get_instance_env(old_site_id, path_to_old_cluster_kubeconfig)
        gitlab_repo = f'viderum/cloud-{old_site_id}'
        if not skip_gitlab:
            CkanGitlab().initialize(gitlab_repo)
        gitlab_registry = f'registry.gitlab.com/{gitlab_repo}'
        old_bucket_name = instance_env.get('CKANEXT__S3FILESTORE__AWS_BUCKET_NAME')
        old_storage_path = instance_env.get('CKANEXT__S3FILESTORE__AWS_STORAGE_PATH')
        assert old_bucket_name == 'ckan'
        assert old_storage_path and len(old_storage_path) > 1
        storage_path = f'/ckan/{old_storage_path}'
        deis_instance_class.create('from-gcloud-envvars', instance_env, gitlab_registry, solr_config,
                                   storage_path, new_instance_id, db_migration_name=db_migration_name)
    routers_env_id = routers_provider.get_env_id()
    default_root_domain = routers_provider.get_default_root_domain()
    assert routers_env_id and default_root_domain
    ckan_site_url = f'https://cc-{routers_env_id}-{new_instance_id}.{default_root_domain}'
    logs.info(f'updating instance and setting ckan site url to {ckan_site_url}', **log_labels)
    deis_instance_class(
        new_instance_id,
        override_spec={
            'envvars': {'CKAN_SITE_URL': ckan_site_url}, **(
                {'db': {'no-db-proxy': 'yes'}, 'datastore': {'no-db-proxy': 'yes'}} if no_db_proxy else {}
            )
        },
        persist_overrides=True
    ).update(wait_ready=True, skip_solr=skip_solr, skip_deployment=skip_deployment)
    if routers_manager.get_deis_instance_routes(new_instance_id):
        logs.info('default instance route already exists', **log_labels)
    else:
        logs.info('creating instance route', **log_labels)
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'deis-instance',
            'deis-instance-id': new_instance_id,
            'root-domain': 'default',
            'sub-domain': 'default'
        })
    if not skip_routes:
        routers_manager.update(router_name, wait_ready=True)
    if not skip_solr:
        logs.info('Rebuilding solr search index', **log_labels)
        deis_instance_class(new_instance_id).ckan.paster('search-index rebuild --force')



import binascii
import os
import yaml
import traceback
import time
import json
import datetime
from pathlib import Path

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers.ckan.deployment import manager as deployment_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.labels import manager as labels_manager

from ckan_cloud_operator.infra import CkanInfra
from ckan_cloud_operator.deis_ckan.migrate import migrate_from_deis

from .db import migration as ckan_db_migration_manager

from .constants import INSTANCE_CRD_SINGULAR, INSTANCE_CRD_PLURAL_SUFFIX, INSTANCE_CRD_KIND_SUFFIX
from .constants import INSTANCE_NAME_CRD_SINGULAR, INSTANCE_NAME_CRD_KIND_SUFFIX, INSTANCE_NAME_CRD_PLURAL_SUFFIX


def initialize(interactive=False):
    ckan_db_migration_manager.initialize(interactive=interactive)
    registry_secrets = config_manager.interactive_set(
        {'disable-centralized-datapushers': 'no'},
        configmap_name='global-ckan-config',
        interactive=interactive
    )
    registry_secrets = config_manager.interactive_set(
        {'private-registry': 'n'},
        secret_name='ckan-docker-registry',
        interactive=interactive
    )
    if registry_secrets.get('private-registry') == 'y':
        registry_secrets = config_manager.interactive_set(
            {
                'docker-server': None,
                'docker-username': None,
                'docker-password': None,
                'docker-email': None,
                'docker-image-pull-secret-name': 'container-registry',
            },
            secret_name='ckan-docker-registry',
            interactive=interactive
        )
    if config_manager.get('enable-deis-ckan', configmap_name='global-ckan-config') == 'y':
        ckan_infra = CkanInfra(required=False)
        config_manager.interactive_set(
            {
                'deis-kubeconfig': ckan_infra.DEIS_KUBECONFIG,
            },
            from_file=True,
            secret_name='ckan-migration-secrets',
            interactive=interactive
        )
        config_manager.interactive_set(
            {
                'gitlab-token': ckan_infra.GITLAB_TOKEN_PASSWORD,
            },
            secret_name='ckan-migration-secrets',
            interactive=interactive
        )
        config_manager.interactive_set(
            {
                'docker-server': ckan_infra.DOCKER_REGISTRY_SERVER,
                'docker-username': ckan_infra.DOCKER_REGISTRY_USERNAME,
                'docker-password': ckan_infra.DOCKER_REGISTRY_PASSWORD,
                'docker-email': ckan_infra.DOCKER_REGISTRY_EMAIL,
            },
            secret_name='ckan-docker-registry',
            interactive=interactive
        )
    crds_manager.install_crd(INSTANCE_CRD_SINGULAR, INSTANCE_CRD_PLURAL_SUFFIX, INSTANCE_CRD_KIND_SUFFIX)
    crds_manager.install_crd(INSTANCE_NAME_CRD_SINGULAR, INSTANCE_NAME_CRD_PLURAL_SUFFIX, INSTANCE_NAME_CRD_KIND_SUFFIX)

    from ckan_cloud_operator.providers.solr.manager import zk_list_configs, zk_put_configs
    logs.info('Checking CKAN Solr config in ZooKeeper')
    if 'ckan_default' in zk_list_configs():
        logs.info('Found ckan_default Solr config')
    else:
        logs.info('No default Solr config found. Putting CKAN 2.8 config for Solr to ZooKeeper as ckan_default...')
        file_path = os.path.dirname(os.path.abspath(__file__))
        root_path = Path(file_path).parent.parent
        configs_dir = os.path.join(root_path, 'data', 'solr')
        zk_put_configs(configs_dir)


    if config_manager.get('disable-centralized-datapushers', configmap_name='global-ckan-config', required=False) != 'yes':
        from ckan_cloud_operator.datapushers import initialize as datapusher_initialize
        datapusher_initialize()

    from ckan_cloud_operator.routers import manager as routers_manager
    from ckan_cloud_operator.providers.routers import manager as router_provider_manager
    router_name = get_default_instances_router_name()
    wildcard_ssl_domain = routers_manager.get_default_root_domain()
    dns_provider = router_provider_manager.get_dns_provider()
    logs.info(f'wildcard_ssl_domain={wildcard_ssl_domain} dns_provider={dns_provider}')
    allow_wildcard_ssl = routers_manager.get_env_id() == 'p'
    router = routers_manager.get(router_name, required=False)
    if router:
        assert (
            (allow_wildcard_ssl and router.get('spec', {}).get('wildcard-ssl-domain') == wildcard_ssl_domain)
            or
            (not allow_wildcard_ssl and not router.get('spec', {}))
        ), f'invalid router wildcard ssl config: {router}'
    else:
        # We don't want to create traefik routes if no dns_provider, unless it's minikube
        if dns_provider.lower() == 'none' and cluster_manager.get_provider_id() != 'minikube':
            pass
        else:
            routers_manager.create(
                router_name,
                routers_manager.get_traefik_router_spec(
                    dns_provider=dns_provider,
                    wildcard_ssl_domain=wildcard_ssl_domain
                )
            )

    from .storage.manager import initialize as ckan_storage_initialize
    ckan_storage_initialize(interactive=interactive)

    from .deployment.manager import initialize as ckan_deployment_initialize
    ckan_deployment_initialize(interactive=interactive)


def get_docker_credentials():
    return [
        config_manager.get(key, secret_name='ckan-docker-registry', required=True)
        for key in ['docker-server', 'docker-username', 'docker-password', 'docker-email']
    ]


def get_default_instances_router_name():
    return 'instances-default'


def migrate_deis_instance(old_site_id, new_instance_id=None, router_name=None, skip_gitlab=False,
                          force=False, rerun=False, recreate_dbs=False, recreate=False, recreate_instance=False,
                          skip_routes=False, skip_solr=False, skip_deployment=False, no_db_proxy=False):
    """Run a full end-to-end migration of an instasnce"""
    from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
    if not router_name:
        router_name = get_default_instances_router_name()
    if not new_instance_id:
        new_instance_id = old_site_id
    if recreate or recreate_instance:
        DeisCkanInstance(new_instance_id).delete(force=True, wait_deleted=True)
    success = False
    db_name = f'{new_instance_id}'
    datastore_name = f'{new_instance_id}-datastore'
    migration_name = None
    for event in ckan_db_migration_manager.migrate_deis_dbs(
            old_site_id, db_name, datastore_name,
            force=force or recreate,
            rerun=rerun or recreate,
            recreate_dbs=recreate_dbs or recreate
    ):
        migration_name = ckan_db_migration_manager.get_event_migration_created_name(event) or migration_name
        success = ckan_db_migration_manager.print_event_exit_on_complete(
            event,
            f'DBs Migration from old site ID: {old_site_id}',
            soft_exit=True
        )
        if success is not None:
            break
    assert success, f'Invalid DB migration success value ({success})'
    logs.info('DB migration completed successfully, continuing with instance migration')
    migrate_from_deis(
        old_site_id, new_instance_id, router_name, DeisCkanInstance,
        skip_gitlab=skip_gitlab, db_migration_name = migration_name,
        recreate=recreate or recreate_instance,
        skip_routes=skip_routes,
        skip_solr=skip_solr,
        skip_deployment=skip_deployment,
        no_db_proxy=no_db_proxy
    )
    post_create_checks(new_instance_id)


def deis_kubeconfig():
    return config_manager.get('deis-kubeconfig', secret_name='ckan-migration-secrets', required=True)


def gitlab_token(token_name=None):
    if token_name:
        return config_manager.get(token_name, secret_name='ckan-gitlab-tokens', required=True)
    else:
        return config_manager.get('gitlab-token', secret_name='ckan-migration-secrets', required=True)


def get_jenkins_token(token_name=None):
    return config_manager.get(secret_name='ckan-jenkins-tokens', key=token_name, required=True).split(',')


def gitlab_search_replace(lines):
    values = config_manager.get('gitlab-search-replace', secret_name='ckan-migration-secrets', required=False)
    if values: values = json.loads(values)
    needs_update = False
    _lines = []
    for line in lines:
        if values:
            for k, v in values.items():
                if k in line:
                    logs.info(f'Search-replace gitlab line: {line} / {k} = {v}')
                    line = line.replace(k, v)
                    needs_update = True
        _lines.append(line)
    return needs_update, _lines


def instance_kind():
    crd_prefix = crds_manager.get_crd_prefix()
    return f'{crd_prefix}{INSTANCE_CRD_KIND_SUFFIX}'


def get_all_dbs_users():
    from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
    from ckan_cloud_operator.providers.db import manager as db_manager
    dbs, users = [], []
    db_host, db_port = db_manager.get_internal_unproxied_db_host_port()
    instance_db_names = []
    instance_user_names = []
    for item in DeisCkanInstance.list(quick=True, return_list=True):
        instance_id = item['id']
        instance = DeisCkanInstance(instance_id)
        spec = instance.spec
        db_name = spec.db['name']
        db_password = instance.annotations.get_secret('databasePassword')
        datastore_name = spec.datastore['name']
        datastore_password = instance.annotations.get_secret('datastorePassword')
        datastore_ro_user = instance.annotations.get_secret('datastoreReadonlyUser')
        datastore_ro_password = instance.annotations.get_secret('datatastoreReadonlyPassword')
        if all([db_name, db_password, datastore_name, datastore_password, datastore_ro_user, datastore_ro_password]):
            dbs.append((db_name, db_host, db_port))
            dbs.append((datastore_name, db_host, db_port))
            instance_db_names += [db_name, datastore_name]
            users.append((db_name, db_password))
            users.append((datastore_name, datastore_password))
            users.append((datastore_ro_user, datastore_ro_password))
            instance_user_names += [db_name, datastore_name, datastore_ro_user]
    migration_dbs, migration_users = ckan_db_migration_manager.get_all_dbs_users()
    for migration_db in migration_dbs:
        if migration_db[0] and migration_db[0] not in instance_db_names:
            dbs.append(migration_db)
    for migration_user in migration_users:
        if migration_user[0] and migration_user[0] not in instance_user_names:
            users.append(migration_user)
    return dbs, users


def get_path_to_old_cluster_kubeconfig():
    path_to_kubeconfig = '/etc/ckan-cloud/viderum-omc/.kube-config'
    if not os.path.exists(path_to_kubeconfig):
        kubeconfig = yaml.load(deis_kubeconfig())
        for filename, content in kubeconfig['__files'].items():
            if not os.path.exists(filename):
                print(f'creating file required for deis kubeconfig: {filename}')
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, 'w') as f:
                    f.write(content)
                print('file created successfully')
        print(f'creating deis kubeconfig: {path_to_kubeconfig}')
        os.makedirs(os.path.dirname(path_to_kubeconfig), exist_ok=True)
        with open(path_to_kubeconfig, 'w') as f:
            f.write(yaml.dump(kubeconfig))
        print('created deis kubeconfig')
    return path_to_kubeconfig


def post_create_checks(instance_id):
    from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
    from ckan_cloud_operator.deis_ckan.envvars import DeisCkanInstanceEnvvars
    instance = DeisCkanInstance(instance_id)
    envvars = DeisCkanInstanceEnvvars(instance).get(full=True)
    assert envvars['ready']
    envvars = envvars['envvars']
    check_envvars(envvars, instance)


def check_envvars(envvars, deis_instance):
    from ckan_cloud_operator.providers.db import manager as db_manager
    from ckan_cloud_operator.deis_ckan.ckan import DeisCkanInstanceCKAN
    logs.info('waiting for instance to be executable...')
    for i in range(20):
        try:
            DeisCkanInstanceCKAN(deis_instance).exec('true')
            break
        except Exception:
            traceback.print_exc()
            logs.warning(f'waiting for instance exec ({i}/20)...')
            time.sleep(5)
            continue
    for url_type, url in {
        'site': envvars.get('CKAN_SITE_URL'),
        'db': envvars.get('CKAN_SQLALCHEMY_URL'),
        'datastore': envvars.get('CKAN__DATASTORE__WRITE_URL'),
        'datastore-ro': envvars.get('CKAN__DATASTORE__READ_URL'),
        'beaker': envvars.get('CKAN___BEAKER__SESSION__URL'),
        'datapusher': envvars.get('CKAN__DATAPUSHER__URL'),
        's3': envvars.get('CKANEXT__S3FILESTORE__HOST_NAME'),
        'solr': envvars.get('CKAN_SOLR_URL'),
    }.items():
        logs.info(f'checking envvar: {url_type}')
        assert url and 'None' not in url, f'invalid url: {url}'
        if url_type in ['site', 'datapusher', 's3']:
            check_cluster_url(url, verify_site_url=url_type == 'site', deis_instance=deis_instance)
        elif url_type in ['db', 'datastore', 'datastore-ro', 'beaker']:
            if '@ckan-cloud-provider-db-proxy-pgbouncer.ckan-cloud:5432' in url:
                db_manager.check_connection_string(
                    url.replace('ckan-cloud-provider-db-proxy-pgbouncer.ckan-cloud', 'localhost'))
        elif url_type == 'solr':
            from ckan_cloud_operator.providers.solr import manager as solr_manager
            assert url.startswith(solr_manager.get_internal_http_endpoint())
        else:
            raise Exception(f'unknown url type: {url_type}')


def check_cluster_url(url, verify_site_url=False, deis_instance=None):
    from ckan_cloud_operator.deis_ckan.ckan import DeisCkanInstanceCKAN
    from ckan_cloud_operator.providers.routers import manager as routers_manager
    env_id = routers_manager.get_env_id()
    root_domain = routers_manager.get_default_root_domain()
    assert url.startswith(f'https://cc-{env_id}-') and url.rstrip('/').endswith(root_domain), f'invalid cluster url: {url}'
    if verify_site_url:
        logs.info('checking site url in running instance...')
        site_url = None
        for line in DeisCkanInstanceCKAN(deis_instance).exec('env', check_output=True).decode().splitlines():
            if line.startswith('CKAN_SITE_URL='):
                site_url = line.replace('CKAN_SITE_URL=', '').strip()
        assert site_url == url, f'mismatch between envvar and instance env: {site_url} != {url}'


def verify_instance_dbs(verify_instance_id):
    from ckan_cloud_operator.providers.db import manager as db_manager
    logs.info(f'{verify_instance_id}: Checking DB..')
    db_manager.check_connection_string(db_manager.get_deis_instsance_external_connection_string(verify_instance_id))
    logs.info(f'{verify_instance_id}: Checking DataStore..')
    db_manager.check_connection_string(db_manager.get_deis_instsance_external_connection_string(verify_instance_id,
                                                                                                is_datastore=True))
    logs.info(f'{verify_instance_id}: Checking DataStore ReadOnly..')
    db_manager.check_connection_string(db_manager.get_deis_instsance_external_connection_string(verify_instance_id,
                                                                                                is_datastore_readonly=True))


def ckan_admin_credentials(instance_id):
    from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
    from ckan_cloud_operator.deis_ckan.ckan import DeisCkanInstanceCKAN
    return DeisCkanInstanceCKAN(DeisCkanInstance(instance_id)).admin_credentials()


def update_deis_instance_envvars(deis_instance, envvars):
    smtp_creds = config_manager.get(secret_name='ckan-default-smtp-credentials')
    old_froms = smtp_creds.get('old-froms')
    old_froms = old_froms.split(',') if old_froms else []
    if smtp_creds and envvars.get('CKAN_SMTP_MAIL_FROM', '') in ['', smtp_creds['from']] + old_froms:
        logs.info(f'updating smtp credentials for deis instance {deis_instance.id}')
        envvars.update(CKAN_SMTP_MAIL_FROM=smtp_creds['from'],
                       CKAN_SMTP_SERVER=smtp_creds['server'],
                       CKAN_SMTP_USER=smtp_creds['user'],
                       CKAN_SMTP_PASSWORD=smtp_creds['password'])
        if envvars.get('CKANEXT__ORGPORTALS__SMTP__MAIL__FROM'):
            logs.info(f'updating orgportls smtp credentials for deis instance {deis_instance.id}')
            envvars.update(CKANEXT__ORGPORTALS__SMTP__MAIL__FROM=smtp_creds['from'],
                           CKANEXT__ORGPORTALS__SMTP__SERVER=smtp_creds['server'],
                           CKANEXT__ORGPORTALS__SMTP__USER=smtp_creds['user'],
                           CKANEXT__ORGPORTALS__SMTP__PASSWORD=smtp_creds['password'])



import json

from ckan_cloud_operator import logs, kubectl
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers.cluster.manager import _generate_password
from ckan_cloud_operator.providers.cluster.aws.manager import aws_check_output, get_aws_credentials, _config_get as aws_config_get

from ..constants import CONFIG_NAME
from .constants import PROVIDER_ID


def initialize(interactive=False, storage_suffix='', use_existing_disk_name=False, dry_run=False):
    default_zone = aws_config_get('aws-default-region', is_secret=True)
    assert default_zone, 'No cluster region specified.'

    if interactive and not dry_run:
        config_manager.interactive_set(
            {'storage-region': default_zone},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )
        config_manager.interactive_set(
            {'aws-storage-access-key': None},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )
        config_manager.interactive_set(
            {'aws-storage-access-secret': None},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )


def create_bucket(instance_id, region=None, exists_ok=False, dry_run=False):
    if not region:
        region = config_manager.get('storage-region') or get_aws_credentials().get('region')

    assert region, 'No default region set for the cluster'

    bucket_exists = instance_id in list_s3_buckets(names_only=True)
    if not exists_ok and bucket_exists:
        raise Exception('Bucket for this instance already exists')

    bucket_name = '{}-cc{}'.format(instance_id, _generate_password(12))

    if not dry_run and not bucket_exists:
        aws_check_output(f's3 mb s3://{bucket_name} --region {region}')

    return {
        'BUCKET_NAME': f's3://{bucket_name}',
        'BUCKET_ACCESS_KEY': config_manager.get('aws-storage-access-key', secret_name=CONFIG_NAME),
        'BUCKET_ACCESS_SECRET': config_manager.get('aws-storage-access-secret', secret_name=CONFIG_NAME)
    }
    

def delete_bucket(instance_id, dry_run=False):
    s3_buckets = list(filter(lambda x: x.startswith(f'{instance_id}-cc'), list_s3_buckets(names_only=True)))
    if not s3_buckets:
        logs.warning(f'No bucket found for the instance "{instance_id}". Skipping.')
        return

    instance = kubectl.get(f'ckancloudckaninstance {instance_id}')
    bucket = instance['spec'].get('ckanStorageBucket').get(PROVIDER_ID)
    if not bucket:
        logs.warning('This instance does not have S3 bucket attached.')
        return

    bucket_name = bucket.get('BUCKET_NAME')

    cmd = f's3 rm {bucket_name} --recursive'
    if dry_run:
        cmd += ' --dryrun'

    # Two steps deletion. See the `aws s3 rb help`
    aws_check_output(cmd)
    if not dry_run:
        aws_check_output(f's3 rb {bucket_name}')


def get_bucket(instance_id):
    s3_buckets = list(filter(lambda x: x.startswith(f'{instance_id}-cc'), list_s3_buckets(names_only=True)))
    if not s3_buckets:
        logs.warning(f'No bucket found for the instance "{instance_id}" on S3. Skipping.')
        return

    instance = kubectl.get(f'ckancloudckaninstance {instance_id}')
    bucket = instance['spec'].get('ckanStorageBucket').get(PROVIDER_ID)
    if not bucket:
        logs.warning('This instance does not have S3 bucket attached.')
        return

    return {
        'instance_id': instance_id,
        'ckanStorageBucket': bucket
    }


def list_buckets():
    """Returns list of buckets attached to CKAN Instances"""
    result = []
    for item in kubectl.get('ckancloudckaninstance').get('items', []):
        bucket = item['spec'].get('ckanStorageBucket', {}).get(PROVIDER_ID)
        if not bucket:
            continue
        result.append({
            'instance_id': item['spec']['id'],
            'ckanStorageBucket': bucket
        })

    return result


def list_s3_buckets(names_only=False):
    """Returns list of buckets on S3"""
    data = aws_check_output('s3 ls').decode()

    result = []
    for line in data.split('\n'):
        if not line.strip():
            continue

        timestamp, name = line.rsplit(maxsplit=1)
        if names_only:
            result.append(name)
        else:
            result.append([timestamp, name])

    return result



from ckan_cloud_operator import kubectl


class CkanRoutersAnnotations(kubectl.BaseAnnotations):
    """Manage router annotations"""

    @property
    def FLAGS(self):
        """Boolean flags which are saved as annotations on the resource"""
        return [
            'forceCreateAnnotations',
            'letsencryptCloudflareEnabled'
        ]

    @property
    def STATUSES(self):
        """Predefined statuses which are saved as annotations on the resource"""
        return {
            'router': ['created']
        }

    @property
    def SECRET_ANNOTATIONS(self):
        """Sensitive details which are saved in a secret related to the resource"""
        return [
            'LETSENCRYPT_CLOUDFLARE_EMAIL',
            'LETSENCRYPT_CLOUDFLARE_API_KEY'
        ]

    @property
    def JSON_ANNOTATION_PREFIXES(self):
        """flexible annotations encoded to json and permitted using key prefixes"""
        return ['default-root-domain']

    @property
    def RESOURCE_KIND(self):
        return 'CkanCloudRouter'

    def get_secret_labels(self):
        return {'ckan-cloud/annotations-secret': self.resource_id,
                'ckan-cloud/router-name': self.resource_id}



import yaml
import copy
from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs


class DeisCkanInstanceSpec(object):

    def __init__(self, spec, override_spec):
        self.spec = copy.deepcopy(spec)
        self.num_applied_overrides = 0
        if override_spec:
            for k, v in override_spec.items():
                if k == 'envvars':
                    logs.info('Applying overrides to instance spec envvars')
                    logs.debug(f"spec['envvars']['overrides'] = {v}")
                    self.spec['envvars']['overrides'] = v
                    self.num_applied_overrides += 1
                elif k in ['db', 'datastore', 'solrCloudCollection']:
                    for kk, vv in v.items():
                        if kk == 'name':
                            print(f'Overriding instance {k} spec name')
                            self.spec.setdefault(k)[kk] = vv
                            self.num_applied_overrides += 1
                        elif k == 'solrCloudCollection' and kk == 'configName':
                            print(f'Overriding instance solr spec config name')
                            self.spec.setdefault(k)[kk] = vv
                            self.num_applied_overrides += 1
                        elif k in ['db', 'datastore'] and kk == 'no-db-proxy':
                            print(f'Overriding instance {k} spec {kk}={vv}')
                            self.spec[k][kk] = vv
                            self.num_applied_overrides += 1
                        else:
                            raise NotImplementedError(f'Unsupported {k} spec override: {kk}={vv}')
                else:
                    raise NotImplementedError(f'Unsupported instance spec override: {k}: {v}')
        self._validate()
        self.envvars = self.spec['envvars']
        self.db = self.spec['db']
        self.datastore = self.spec['datastore']
        self.solrCloudCollection = self.spec['solrCloudCollection']
        self.storage = self.spec['storage']

    def _validate(self):
        spec = self.spec
        for k, v in spec.items():
            if k == 'ckanPodSpec':
                assert type(v) == dict
            elif k == 'ckanContainerSpec':
                assert type(v) == dict
                if v.get('image'):
                    assert not v.get('imageFromGitlab')
                elif v.get('imageFromGitlab'):
                    assert not v.get('image')
                else:
                    raise ValueError(f'Invalid ckanContainerSpec: {v}')
            elif k in ['db', 'datastore']:
                assert type(v) == dict
                assert v['name']
                for kk, vv in v.items():
                    if kk == 'name':
                        assert type(vv) == str
                    elif kk == 'importGcloudSqlDumpUrl':
                        assert type(vv) == str
                    elif kk == 'fromDeisInstance':
                        assert type(vv) == str
                    elif kk == 'fromDbMigration':
                        assert type(vv) == str
                    elif kk == 'no-db-proxy':
                        assert vv in ['yes', 'no', ''], 'only valid values for no-db-proxy are: "yes", "no", ""'
                    elif kk == 'dbPrefix':
                        assert not vv or type(vv) == str
                    else:
                        raise ValueError(f'Invalid db spec attribute: {kk}={vv}')
            elif k == 'solrCloudCollection':
                assert type(v) == dict
                assert v['name']
                for kk, vv in v.items():
                    if kk == 'name':
                        assert type(vv) == str
                    elif kk == 'configName':
                        assert type(vv) == str and not v.get('configFromDeisInstance')
                    elif kk == 'configFromDeisInstance':
                        assert type(vv) == str and not v.get('configName')
                    else:
                        raise ValueError(f'Invalid solr cloud collection spec attribute: {kk}={vv}')
            elif k == 'envvars':
                assert type(v) == dict
                for kk, vv in v.items():
                    if kk == 'fromSecret':
                        assert type(vv) == str and not v.get('fromGitlab')
                    elif kk == 'overrides':
                        assert type(vv) == dict
                    elif kk == 'fromGitlab':
                        assert type(vv) == str and not v.get('fromSecret')
                    else:
                        raise ValueError(f'Invalid envvars spec attribute: {kk}={vv}')
                assert v.get('fromSecret') or v.get('fromGitlab')
            elif k == 'ckan':
                assert type(v) == dict
                for kk, vv in v.items():
                    if kk == 'init':
                        assert type(vv) == list
                    else:
                        raise ValueError(f'Invalid ckan spec attribute: {kk}={vv}')
            elif k == 'storage':
                assert type(v) == dict
                for kk, vv in v.items():
                    if kk == 'path':
                        assert type(vv) == str and len(vv) > 6, f'Invalid storage path: {vv}'
                    else:
                        raise ValueError(f'Invalid storage spec attribute: {kk}={vv}')
            elif k == 'routes':
                assert type(v) == dict
                for kk, vv in v.items():
                    if kk == 'target-port':
                        assert type(vv) == int and vv > 0, f'invalid target-port: {vv}'
                    else:
                        raise ValueError(f'Invalid routers spec attribute: {kk}={vv}')
            elif k == 'imagePullSecrets':
                assert type(v) == list
            else:
                raise ValueError(f'Invalid spec attribute: {k}={v}')
        assert spec['db']['name'] != spec['datastore']['name']



import datetime

from urllib3.util import parse_url

from ckan_cloud_operator import kubectl

from .. import manager as solr_manager


def deploy():
    """Deploys a proxy inside the cluster which allows to access the centralized solr without authentication"""
    labels = {'app': 'ckan-cloud-solrcloud-proxy'}
    solr_url = parse_url(solr_manager.get_internal_http_endpoint())
    scheme = solr_url.scheme
    hostname = solr_url.hostname
    port = solr_url.port
    solr_user, solr_password = '', ''
    if solr_url.auth:
        solr_user, solr_password = solr_url.auth.split(':')
    if not port:
        port = '443' if scheme == 'https' else '8983'
    kubectl.update_secret('solrcloud-proxy', {
        'SOLR_URL': f'{scheme}://{hostname}:{port}',
        'SOLR_USER': solr_user,
        'SOLR_PASSWORD': solr_password
    })
    kubectl.apply(kubectl.get_deployment('solrcloud-proxy', labels, {
        'replicas': 1,
        'revisionHistoryLimit': 10,
        'strategy': {'type': 'RollingUpdate', },
        'template': {
            'metadata': {
                'labels': labels,
                'annotations': {
                    'ckan-cloud/operator-timestamp': str(datetime.datetime.now())
                }
            },
            'spec': {
                'containers': [
                    {
                        'name': 'solrcloud-proxy',
                        'image': 'viderum/ckan-cloud-operator-solrcloud-proxy',
                        'envFrom': [{'secretRef': {'name': 'solrcloud-proxy'}}],
                        'ports': [{'containerPort': 8983}],
                    }
                ]
            }
        }
    }))
    service = kubectl.get_resource('v1', 'Service', 'solrcloud-proxy', labels)
    service['spec'] = {
        'ports': [
            {'name': '8983', 'port': 8983}
        ],
        'selector': labels
    }
    kubectl.apply(service)




import toml
import time
import hashlib
import json

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.routers.traefik import config as traefik_router_config
from ckan_cloud_operator.routers.routes import manager as routes_manager
from ckan_cloud_operator import cloudflare
from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.labels import manager as labels_manager
from ckan_cloud_operator.config import manager as config_manager


def _get_deployment_spec(router_name, router_type, annotations, image=None, httpauth_secrets=None, dns_provider=None):
    volume_spec = cluster_manager.get_or_create_multi_user_volume_claim(get_label_suffixes(router_name, router_type))
    httpauth_secrets_volume_mounts, httpauth_secrets_volumes = [], []
    if httpauth_secrets:
        added_secrets = []
        for httpauth_secret in httpauth_secrets:
            if httpauth_secret in added_secrets: continue
            added_secrets.append(httpauth_secret)
            httpauth_secrets_volumes.append({
                'name': httpauth_secret, 'secret': {'secretName': httpauth_secret}
            })
            httpauth_secrets_volume_mounts.append({
                'name': httpauth_secret, 'mountPath': f'/httpauth-{httpauth_secret}'
            })
    container_spec_overrides = config_manager.get(
        'container-spec-overrides',
        configmap_name=f'traefik-router-{router_name}-deployment',
        required=False,
        default=None
    )
    deployment_spec = {
        'replicas': 1,
        'revisionHistoryLimit': 5,
        'selector': {
            'matchLabels': get_labels(router_name, router_type, for_deployment=True)
        },
        'template': {
            'metadata': {
                'labels': get_labels(router_name, router_type, for_deployment=True)
            },
            'spec': {
                'containers': [
                    {
                        'name': 'traefik',
                        'image': image or 'traefik:1.6-alpine',
                        'ports': [{'containerPort': 80}],
                        'volumeMounts': [
                            {'name': 'etc-traefik', 'mountPath': '/etc-traefik'},
                            {'name': 'traefik-acme', 'mountPath': '/traefik-acme', 'subPath': f'router-traefik-{router_name}'},
                            *httpauth_secrets_volume_mounts,
                        ],
                        'args': ['--configFile=/etc-traefik/traefik.toml'],
                        **(json.loads(container_spec_overrides) if container_spec_overrides else {})
                    }
                ],
                'volumes': [
                    {'name': 'etc-traefik', 'configMap': {'name': f'router-traefik-{router_name}'}},
                    dict(volume_spec, name='traefik-acme'),
                    *httpauth_secrets_volumes,
                ]
            }
        }
    }
    if dns_provider == 'route53':
        logs.info('Traefik deployment: adding SSL support using AWS Route53')
        container = deployment_spec['template']['spec']['containers'][0]
        container['ports'].append({'containerPort': 443})
        aws_credentials = cluster_manager.get_provider().get_aws_credentials()
        secret_name = f'ckancloudrouter-{router_name}-route53'
        kubectl.update_secret(secret_name, {
            'AWS_ACCESS_KEY_ID': aws_credentials['access'],
            'AWS_SECRET_ACCESS_KEY': aws_credentials['secret'],
            'AWS_REGION': aws_credentials['region']
        }, labels=get_labels(router_name, router_type))
        container['envFrom'] = [{'secretRef': {'name': secret_name}}]
    elif annotations.get_flag('letsencryptCloudflareEnabled'):
        logs.info('Traefik deployment: adding SSL support using Cloudflare')
        container = deployment_spec['template']['spec']['containers'][0]
        container['ports'].append({'containerPort': 443})
        cloudflare_email, cloudflare_api_key = get_cloudflare_credentials()
        secret_name = f'ckancloudrouter-{router_name}-cloudflare'
        kubectl.update_secret(secret_name, {
            'CLOUDFLARE_EMAIL': cloudflare_email,
            'CLOUDFLARE_API_KEY': cloudflare_api_key,
        }, labels=get_labels(router_name, router_type))
        container['envFrom'] = [{'secretRef': {'name': secret_name}}]
    elif dns_provider == 'azure':
        logs.info('Traefik deployment: adding SSL support using Azure DNS')
        container = deployment_spec['template']['spec']['containers'][0]
        container['ports'].append({'containerPort': 443})
        azure_credendials = cluster_manager.get_provider().get_azure_credentials()
        secret_name = f'ckancloudrouter-{router_name}-azure'
        kubectl.update_secret(secret_name, {
            'AZURE_CLIENT_ID': azure_credendials['azure-client-id'],
            'AZURE_CLIENT_SECRET':  azure_credendials['azure-client-secret'],
            'AZURE_SUBSCRIPTION_ID': azure_credendials['azure-subscribtion-id'],
            'AZURE_TENANT_ID': azure_credendials['azure-tenant-id'],
            'AZURE_RESOURCE_GROUP': azure_credendials['azure-resource-group']
        }, labels=get_labels(router_name, router_type))
        container['envFrom'] = [{'secretRef': {'name': secret_name}}]
    else:
        logs.info('Not configuring SSL support for Traefik deployment')
    return deployment_spec


def _get_resource_name(router_name):
    return f'router-traefik-{router_name}'


def _update(router_name, spec, annotations, routes):
    dns_provider = spec.get('dns-provider', 'cloudflare')
    if dns_provider == 'none':
        logs.info('No DNS provider, not setting up ingress')
        return
    resource_name = _get_resource_name(router_name)
    router_type = spec['type']
    cloudflare_email, cloudflare_auth_key = get_cloudflare_credentials()
    external_domains = spec.get('external-domains')
    logs.info('updating traefik deployment', resource_name=resource_name, router_type=router_type,
              cloudflare_email=cloudflare_email, cloudflare_auth_key_len=len(cloudflare_auth_key) if cloudflare_auth_key else 0,
              external_domains=external_domains, dns_provider=dns_provider)
    kubectl.apply(kubectl.get_configmap(
        resource_name, get_labels(router_name, router_type),
        {'traefik.toml': toml.dumps(traefik_router_config.get(
            routes, cloudflare_email,
            enable_access_log=bool(spec.get('enable-access-log')),
            wildcard_ssl_domain=spec.get('wildcard-ssl-domain'),
            external_domains=external_domains,
            dns_provider=dns_provider,
            force=True
        ))}
    ))
    domains = {}
    httpauth_secrets = []
    for route in routes:
        root_domain, sub_domain = routes_manager.get_domain_parts(route)
        domains.setdefault(root_domain, []).append(sub_domain)
        routes_manager.pre_deployment_hook(route, get_labels(router_name, router_type))
        if route['spec'].get('httpauth-secret') and route['spec']['httpauth-secret'] not in httpauth_secrets:
            httpauth_secrets.append(route['spec']['httpauth-secret'])
    load_balancer = kubectl.get_resource(
        'v1', 'Service', f'loadbalancer-{resource_name}',
        get_labels(router_name, router_type)
    )
    load_balancer['spec'] = {
        'ports': [
            {'name': '80', 'port': 80},
            {'name': '443', 'port': 443},
        ],
        'selector': {
            'app': get_labels(router_name, router_type, for_deployment=True)['app']
        },
        'type': 'LoadBalancer'
    }
    kubectl.apply(load_balancer)
    load_balancer_ip = get_load_balancer_ip(router_name)
    logs.info(f'load balancer ip: {load_balancer_ip}')
    from ckan_cloud_operator.providers.routers import manager as routers_manager
    if external_domains:
        from ckan_cloud_operator.providers.routers import manager as routers_manager
        external_domains_router_root_domain = routers_manager.get_default_root_domain()
        env_id = routers_manager.get_env_id()
        assert router_name.startswith('prod-'), f'invalid external domains router name: {router_name}'
        external_domains_router_sub_domain = f'cc-{env_id}-{router_name}'
        routers_manager.update_dns_record(
            dns_provider, external_domains_router_sub_domain, external_domains_router_root_domain,
            load_balancer_ip, cloudflare_email, cloudflare_auth_key
        )
    else:
        for root_domain, sub_domains in domains.items():
            for sub_domain in sub_domains:
                routers_manager.update_dns_record(
                    dns_provider, sub_domain, root_domain,
                    load_balancer_ip, cloudflare_email, cloudflare_auth_key
                )
    kubectl.apply(kubectl.get_deployment(
        resource_name, get_labels(router_name, router_type, for_deployment=True),
        _get_deployment_spec(
            router_name, router_type, annotations,
            image=('traefik:1.7' if (external_domains or len(httpauth_secrets) > 0) else None),
            httpauth_secrets=httpauth_secrets,
            dns_provider=dns_provider
        )
    ))


def get_load_balancer_ip(router_name, failfast=False):
    resource_name = _get_resource_name(router_name)
    RETRIES = 10
    for retries in range(RETRIES):
        load_balancer = kubectl.get(f'service loadbalancer-{resource_name}', required=False)

        if load_balancer:
            ingresses = load_balancer.get('status', {}).get('loadBalancer', {}).get('ingress', [])
            if len(ingresses) > 0:
                assert len(ingresses) == 1
                if cluster_manager.get_provider_id() == 'aws':
                    load_balancer_hostname = ingresses[0].get('hostname')
                    if load_balancer_hostname:
                        return load_balancer_hostname
                    logs.warning('Failed to get hostname, retrying %r' % ingresses[0])
                else:
                    load_balancer_ip = ingresses[0].get('ip')
                    if load_balancer_ip:
                        return load_balancer_ip
                    logs.warning('Failed to get ip, retrying %r' % ingresses[0])
        if failfast:
            return None
        else:
            time.sleep(60)
        assert retries < RETRIES - 1, "Gave up on waiting for load balancer IP"


def get_cloudflare_credentials():
    from ckan_cloud_operator.providers.routers import manager as routers_manager
    cloudflare_email, cloudflare_auth_key = routers_manager.get_cloudflare_credentials()
    return cloudflare_email, cloudflare_auth_key


def update(router_name, wait_ready, spec, annotations, routes, dry_run=False):
    old_deployment = kubectl.get(f'deployment router-traefik-{router_name}', required=False)
    old_generation = old_deployment.get('metadata', {}).get('generation') if old_deployment else None
    expected_new_generation = old_generation + 3 if old_generation else None
    if expected_new_generation:
        print(f'old deployment generation: {old_generation}')
    else:
        print('Creating new deployment')
    if not dry_run:
        annotations.update_status(
            'router', 'created',
            lambda: _update(router_name, spec, annotations, routes),
            force_update=True
        )
        if expected_new_generation:
            _scale_down_scale_up(f'router-traefik-{router_name}')
            while True:
                time.sleep(.2)
                new_deployment = kubectl.get(f'deployment router-traefik-{router_name}', required=False)
                if not new_deployment: continue
                new_generation = new_deployment.get('metadata', {}).get('generation')
                if not new_generation: continue
                if new_generation == old_generation: continue
                if new_generation != expected_new_generation:
                    raise Exception(f'Invalid generation: {new_generation} (expected: {expected_new_generation})')
                print(f'new deployment generation: {new_generation}')
                break
        if wait_ready:
            print('Waiting for instance to be ready...')
            while time.sleep(2):
                if get(router_name)['ready']: break
                print('.')


def get(router_name):
    deployment = kubectl.get(f'deployment/router-traefik-{router_name}', required=False)
    if deployment:
        return kubectl.get_deployment_detailed_status(
            deployment, f'ckan-cloud/router-name={router_name}', 'traefik'
        )
    else:
        return {'ready': False}


def get_dns_data(router_name, router, failfast=False):
    external_domains = router['spec'].get('external-domains')
    data = {
        'load-balancer-ip': get_load_balancer_ip(router_name, failfast=failfast),
    }
    if external_domains:
        from ckan_cloud_operator.providers.routers import manager as routers_manager
        external_domains_router_root_domain = routers_manager.get_default_root_domain()
        env_id = routers_manager.get_env_id()
        assert router_name.startswith('prod-'), f'invalid external domains router name: {router_name}'
        external_domains_router_sub_domain = f'cc-{env_id}-{router_name}'
        data['external-domain'] = f'{external_domains_router_sub_domain}.{external_domains_router_root_domain}'
    return data


def get_label_suffixes(router_name, router_type):
    return {
        'router-name': router_name,
        'router-type': router_type
    }


def get_labels(router_name, router_type, for_deployment=False):
    label_prefix = labels_manager.get_label_prefix()
    extra_labels = {'app': f'{label_prefix}-router-{router_name}'} if for_deployment else {}
    return labels_manager.get_resource_labels(
        get_label_suffixes(router_name, router_type),
        extra_labels=extra_labels
    )

def _scale_down_scale_up(deployment='router-traefik-instances-default', replicas=1):
    kubectl.call(f'scale deployment {deployment} --replicas=0')
    kubectl.call(f'scale deployment {deployment} --replicas={replicas}')



import subprocess
import yaml
from ckan_cloud_operator import kubectl


class DeisCkanInstanceNamespace(object):

    def __init__(self, instance):
        self.instance = instance

    def update(self):
        self.instance.annotations.update_status('namespace', 'created',
                                                lambda: self._initialize_instance_namespace())

    def delete(self):
        ns = self.instance.id
        print(f'Deleting instance namespace: {ns}')
        subprocess.check_call(
            f'kubectl delete -n {ns} rolebinding/ckan-{ns}-operator-rolebinding '
            f'                       role/ckan-{ns}-operator-role '
            f'                       serviceaccount/ckan-{ns}-operator',
            shell=True
        )
        subprocess.check_call(f'kubectl delete ns {ns}', shell=True)

    def get(self):
        exitcode, output = subprocess.getstatusoutput(f'kubectl -n {self.instance.id} get ns/{self.instance.id} -o yaml')
        events = [
            '{firstTimestamp} - {lastTimestamp} ({type}*{count}) {message}'.format(**e)
            for e in kubectl.get('events', namespace=self.instance.id)['items']
            if e['type'] != 'Normal'
        ]
        if exitcode == 0:
            return {'ready': yaml.load(output).get('status', {}).get('phase') == 'Active', 'events': events}
        else:
            return {'ready': False, 'error': output, 'events': events}

    def _initialize_instance_namespace(self):
        ns = self.instance.id
        print(f'initializing instance namespace: {ns}')
        subprocess.check_call(f'kubectl create ns {ns}', shell=True)
        kubectl_namespace = f'kubectl --namespace {ns}'
        subprocess.check_call(f'{kubectl_namespace} create serviceaccount ckan-{ns}-operator',
                              shell=True)
        subprocess.check_call(f'{kubectl_namespace} create role ckan-{ns}-operator-role '
                              f' --verb list,get,create --resource secrets,pods,pods/exec,pods/portforward',
                              shell=True)
        subprocess.check_call(f'{kubectl_namespace} create rolebinding ckan-{ns}-operator-rolebinding'
                              f' --role ckan-{ns}-operator-role'
                              f' --serviceaccount {ns}:ckan-{ns}-operator',
                              shell=True)



import json
import re

from ckan_cloud_operator import logs, kubectl
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers.cluster.gcloud.manager import _config_get as cluster_config_get, check_output as gcloud_check_output

from ..constants import CONFIG_NAME
from .constants import PROVIDER_ID


def initialize(interactive=False, storage_suffix='', use_existing_disk_name=False, dry_run=False):
    default_zone = cluster_config_get('cluster-compute-zone')
    assert default_zone, 'No cluster region specified.'

    if interactive and not dry_run:
        config_manager.interactive_set(
            {'storage-region': default_zone},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )


def create_bucket(instance_id, region=None, exists_ok=False, dry_run=False):
    if not region:
        region = cluster_config_get('cluster-compute-zone')

    assert region, 'No default region set for the cluster'

    bucket_exists = instance_id in list_gcloud_buckets()
    if not exists_ok and bucket_exists:
        raise Exception('Bucket for this instance already exists')

    if not dry_run and not bucket_exists:
        gcloud_check_output(f'mb gs://{instance_id} -l {region}', gsutil=True)

    return {
        'BUCKET_NAME': f'gs://{instance_id}'
    }
    

def delete_bucket(instance_id, dry_run=False):
    if instance_id not in list_gcloud_buckets():
        logs.warning(f'No bucket found for the instance "{instance_id}". Skipping.')
        return

    logs.info(f'Removing bucket for instance_id {instance_id}')
    if not dry_run:
        gcloud_check_output(f'rm -r gs://{instance_id}', gsutil=True)


def get_bucket(instance_id):
    if instance_id not in list_gcloud_buckets():
        logs.warning(f'No bucket found for the instance "{instance_id}" on Google Cloud. Skipping.')
        return

    instance = kubectl.get(f'ckancloudckaninstance {instance_id}')
    bucket = instance['spec'].get('ckanStorageBucket').get(PROVIDER_ID)
    if not bucket:
        logs.warning('This instance does not have Google Cloud bucket attached.')
        return

    return {
        'instance_id': instance_id,
        'ckanStorageBucket': bucket
    }


def list_buckets():
    """Returns list of buckets attached to CKAN Instances"""
    result = []
    for item in kubectl.get('ckancloudckaninstance').get('items', []):
        bucket = item['spec'].get('ckanStorageBucket', {}).get(PROVIDER_ID)
        if not bucket:
            continue
        result.append({
            'instance_id': item['spec']['id'],
            'ckanStorageBucket': bucket
        })

    return result


def list_gcloud_buckets():
    """Returns list of buckets on GCP"""
    data = gcloud_check_output('ls', gsutil=True).decode()

    result = []
    for line in data.split('\n'):
        if not line.strip():
            continue

        bucket_name = re.match(r'gs://([-_a-zA-Z0-9]+)/', line)
        if not bucket_name:
            continue

        result.append(bucket_name.group(1))

    return result



DATASTORE_PERMISSIONS_SQL_TEMPLATE = """
CREATE OR REPLACE VIEW "_table_metadata" AS
    SELECT DISTINCT
        substr(md5(dependee.relname || COALESCE(dependent.relname, '')), 0, 17) AS "_id",
        dependee.relname AS name,
        dependee.oid AS oid,
        dependent.relname AS alias_of
    FROM
        pg_class AS dependee
        LEFT OUTER JOIN pg_rewrite AS r ON r.ev_class = dependee.oid
        LEFT OUTER JOIN pg_depend AS d ON d.objid = r.oid
        LEFT OUTER JOIN pg_class AS dependent ON d.refobjid = dependent.oid
    WHERE
        (dependee.oid != dependent.oid OR dependent.oid IS NULL) AND
        -- is a table (from pg_tables view definition)
        -- or is a view (from pg_views view definition)
        (dependee.relkind = 'r'::"char" OR dependee.relkind = 'v'::"char")
        AND dependee.relnamespace = (
            SELECT oid FROM pg_namespace WHERE nspname='public')
    ORDER BY dependee.oid DESC;
ALTER VIEW "_table_metadata" OWNER TO "{{SITE_USER}}";
GRANT SELECT ON "_table_metadata" TO "{{DS_RO_USER}}";

CREATE OR REPLACE FUNCTION populate_full_text_trigger() RETURNS trigger
AS $body$
    BEGIN
        IF NEW._full_text IS NOT NULL THEN
            RETURN NEW;
        END IF;
        NEW._full_text := (
            SELECT to_tsvector(string_agg(value, ' '))
            FROM json_each_text(row_to_json(NEW.*))
            WHERE key NOT LIKE '\_%');
        RETURN NEW;
    END;
$body$ LANGUAGE plpgsql;
ALTER FUNCTION populate_full_text_trigger() OWNER TO "{{SITE_USER}}";

DO $body$
    BEGIN
        EXECUTE coalesce(
            (SELECT string_agg(
                'CREATE TRIGGER zfulltext BEFORE INSERT OR UPDATE ON ' ||
                quote_ident(relname) || ' FOR EACH ROW EXECUTE PROCEDURE ' ||
                'populate_full_text_trigger();', ' ')
            FROM pg_class
            LEFT OUTER JOIN pg_trigger AS t
                ON t.tgrelid = relname::regclass AND t.tgname = 'zfulltext'
            WHERE relkind = 'r'::"char" AND t.tgname IS NULL
                AND relnamespace = (
                    SELECT oid FROM pg_namespace WHERE nspname='public')),
            'SELECT 1;');
    END;
$body$;
"""



import click
import datetime
import traceback
import subprocess

from ckan_cloud_operator import logs

from . import manager

from .gcloudsql import cli as gcloudsql_proxy_cli
from ckan_cloud_operator.config import manager as config_manager


@click.group()
def proxy():
    """Manage the centralized db proxy"""
    pass


proxy.add_command(gcloudsql_proxy_cli.gcloudsql)


@proxy.command()
@click.option('--db-prefix')
@click.option('--all-daemon')
def port_forward(db_prefix, all_daemon):
    if all_daemon:
        assert not db_prefix and all_daemon == 'I know the risks'
        subprocess.Popen(['ckan-cloud-operator', 'db', 'proxy', 'port-forward'], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        for db_prefix in manager.get_provider().get_all_db_prefixes():
            subprocess.Popen(['ckan-cloud-operator', 'db', 'proxy', 'port-forward', '--db-prefix', db_prefix],
                             stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
    else:
        while True:
            start_time = datetime.datetime.now()
            try:
                manager.start_port_forward(db_prefix=db_prefix)
            except Exception:
                traceback.print_exc()
            end_time = datetime.datetime.now()
            if (end_time - start_time).total_seconds() < 10:
                logs.critical('DB Proxy failure')
                logs.exit_catastrophic_failure()
            else:
                logs.warning('Restarting the DB proxy')


@proxy.command()
@click.option('--verify-instance-id')
@click.option('--set-pool-mode')
@click.option('--reload', is_flag=True)
def proxy_update(verify_instance_id, set_pool_mode, reload):
    manager.update(wait_updated=reload, set_pool_mode=set_pool_mode)
    if verify_instance_id:
        from ckan_cloud_operator.providers.db import manager as db_manager
        logs.info(f'{verify_instance_id}: Checking DB..')
        db_manager.check_connection_string(db_manager.get_deis_instsance_external_connection_string(verify_instance_id))
        logs.info(f'{verify_instance_id}: Checking DataStore..')
        db_manager.check_connection_string(db_manager.get_deis_instsance_external_connection_string(verify_instance_id,
                                                                                                    is_datastore=True))
        logs.info(f'{verify_instance_id}: Checking DataStore ReadOnly..')
        db_manager.check_connection_string(db_manager.get_deis_instsance_external_connection_string(verify_instance_id,
                                                                                                    is_datastore_readonly=True))
    logs.exit_great_success()


@proxy.command()
def initialize():
    manager.initialize()
    logs.exit_great_success()



import click
import yaml

from . import driver


@click.group()
def kubectl():
    """Manage Kubernetes resources unrelated to the operator or cluster"""
    pass


@kubectl.command()
@click.argument('WHAT')
@click.argument('ARGS', nargs=-1)
@click.option('--required', is_flag=True)
@click.option('--namespace')
@click.option('--get-cmd', default='get')
@click.option('--default-flow-style', is_flag=True)
def get(what, args, required, namespace, get_cmd, default_flow_style):
    if not namespace: namespace = 'ckan-cloud'
    print(yaml.dump(
        driver.get(what, *args, required=required, namespace=namespace, get_cmd=get_cmd),
        default_flow_style=default_flow_style
    ))



PROVIDER_SUBMODULE='ckan-db'

MIGRATION_CRD_SINGULAR = 'dbmigration'
MIGRATION_CRD_PLURAL = 'dbmigrations'
MIGRATION_CRD_KIND = 'DbMigration'

MIGRATION_EVENT_STATUS_SUCCESS = 'success'
MIGRATION_EVENT_STATUS_FAILURE = 'failure'



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from ..constants import PROVIDER_SUBMODULE
from .constants import PROVIDER_ID

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)

################################

# custom provider code starts here
from ckan_cloud_operator import kubectl
from ckan_cloud_operator.providers.db import manager as db_manager


def initialize():
    update()
    _set_provider()


def update():
    _apply_deployment()


def start():
    print('\n'.join([
        '',
        '\nadmin db credentials:\n' + str(db_manager.get_admin_db_credentials()),
        '\ninternal proxy host / port:\n' + str(db_manager.get_internal_proxy_host_port()),
        '\ninternal unproxied host / port:\n' + str(db_manager.get_internal_unproxied_db_host_port()),
        '\n'
    ]))
    deployment_name = _get_resource_name()
    kubectl.check_call(f'port-forward deployment/{deployment_name} 8080')


def _apply_deployment():
    deployment_name = _get_resource_name()
    deployment_labels = _get_resource_labels(for_deployment=True)
    deployment_annotations = _get_resource_annotations()
    kubectl.apply(kubectl.get_deployment(
        deployment_name,
        deployment_labels,
        {
            'replicas': 1,
            'revisionHistoryLimit': 2,
            'strategy': {'type': 'RollingUpdate', },
            'template': {
                'metadata': {
                    'labels': deployment_labels,
                    'annotations': deployment_annotations,
                },
                'spec': {
                    'containers': [
                        {
                            'name': 'adminer',
                            'image': 'adminer',
                            'ports': [{'containerPort': 8080}],
                        }
                    ],
                }
            }
        }
    ))



import click
import yaml

from ckan_cloud_operator import logs

from .proxy import cli as proxy_cli
from . import manager


@click.group()
def solr():
    """Manage SOLR"""
    pass


solr.add_command(proxy_cli.proxy)


@solr.command()
@click.option('--interactive', is_flag=True)
@click.option('--dry-run', is_flag=True)
def initialize(interactive, dry_run):
    manager.initialize(interactive=interactive, dry_run=dry_run)
    logs.exit_great_success()


@solr.command()
def zoonavigator_port_forward():
    manager.start_zoonavigator_port_forward()


@solr.command()
@click.option('--suffix', default='sc-3')
def solrcloud_port_forward(suffix):
    manager.start_solrcloud_port_forward(suffix)


@solr.command()
@click.argument('COLLECTION_NAME')
def collection_status(collection_name):
    logs.print_yaml_dump(manager.get_collection_status(collection_name))


@solr.command()
@click.argument('PATH')
def curl(path):
    print(manager.solr_curl(path, required=True, debug=True))


@solr.command()
def internal_http_endpoint():
    print(manager.get_internal_http_endpoint())


#### ZooKeeper


@click.group()
def zk():
    """Manage ZooKeeper"""
    pass


@zk.command()
@click.option('--config-name')
@click.option('--output-dir')
@click.option('--filename')
@click.option('--all', is_flag=True)
def get_configs(config_name, output_dir, filename, all):
    if config_name:
        config_names = [config_name]
    else:
        config_names = manager.zk_list_configs()
        if not all:
            config_names = [cn for cn in config_names if 'ckan' in cn]
    for zk_config_name in config_names:
        print(f'-- {zk_config_name}')
        if filename:
            config_files = [filename]
        else:
            config_files = []
            manager.zk_list_config_files(zk_config_name, config_files)
        for zk_filename in config_files:
            print(f'/{zk_config_name}{zk_filename}')
            if output_dir:
                output_filename = f'{output_dir}/{zk_config_name}{zk_filename}'
                manager.zk_get_config_file(zk_config_name, zk_filename, output_filename)
                print(f'--> {output_filename}')


@zk.command()
@click.argument('CONFIGS_DIR')
def put_configs(configs_dir):
    manager.zk_put_configs(configs_dir)
    logs.exit_great_success()


solr.add_command(zk)



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import yaml
import json
import subprocess
import datetime
import os
import binascii
import collections

from ckan_cloud_operator import kubectl
from ckan_cloud_operator.drivers.kubectl import rbac
from ckan_cloud_operator import logs


def initialize(interactive=False):
    _set_provider()
    if interactive:
        print('\nEnter credentials for an AWS Access Key with relevant permissions\n')
        _config_interactive_set({'aws-access-key-id': None}, is_secret=True)
        _config_interactive_set({'aws-secret-access-key': None}, is_secret=True)
        print('\nEnter the AWS Region the Amazon Kubernets cluster is hosted on\n')
        _config_interactive_set({'aws-default-region': None}, is_secret=True)
        print('\nEnter the name of your Amazon EKS cluster\n')
        _config_interactive_set({'eks-cluster-name': None}, is_secret=True)
    print(yaml.dump(get_info(), default_flow_style=False))
    from ckan_cloud_operator.providers.storage.efs import manager as efs_manager
    efs_manager.initialize(interactive=interactive)
    _create_storage_classes()
    _update_service_account()


def _update_service_account():
    rbac.update_cluster_role_binding(
        name='default-admin-rbac',
        subject=dict(
            kind='ServiceAccount',
            name='default',
            namespace='default',
        ),
        cluster_role_name='cluster-admin',
        labels=_get_resource_labels()
    )


def _create_storage_classes():
    kubectl.apply({
        'apiVersion': 'storage.k8s.io/v1', 'kind': 'StorageClass',
        'metadata': {
            'name': 'cca-ckan',
        },
        'provisioner': 'example.com/aws-efs',
        'reclaimPolicy': 'Delete',
        'volumeBindingMode': 'Immediate'
    })
    kubectl.apply({
        'apiVersion': 'storage.k8s.io/v1', 'kind': 'StorageClass',
        'metadata': {
            'name': 'cca-storage',
        },
        'provisioner': 'kubernetes.io/aws-ebs',
        'reclaimPolicy': 'Delete',
        'volumeBindingMode': 'Immediate',
        'parameters': {
            'encrypted': 'false',
            'type': 'gp2',
        }
    })


def get_info(debug=False):
    cluster = _config_get('eks-cluster-name', is_secret=True)
    data = yaml.load(aws_check_output(f'eks describe-cluster --name {cluster}'))['cluster']
    if debug:
        return data
    else:
        return {
            'name': data['name'],
            'status': data['status'],
            # 'zone': data['zone'],
            # 'locations': data['locations'],
            'endpoint': data['endpoint'],
            # 'nodePools': [
            #     {
            #         'name': pool['name'],
            #         'status': pool['status'],
            #         'version': pool['version'],
            #         'config': {
            #             'diskSizeGb': pool['config']['diskSizeGb'],
            #             'machineType': pool['config']['machineType'],
            #         },
            #     } for pool in data['nodePools']
            # ],
            'createTime': datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc) + datetime.timedelta(seconds=float(data['createdAt'])),
            'currentMasterVersion': data['version'],
            # 'currentNodeCount': data['currentNodeCount'],
        }


def get_aws_credentials():
    return {
        'access': _config_get('aws-access-key-id', is_secret=True),
        'secret': _config_get('aws-secret-access-key', is_secret=True),
        'region': _config_get('aws-default-region', is_secret=True)
    }

def aws_process_cmd(cmd):
    access = _config_get('aws-access-key-id', is_secret=True)
    secret = _config_get('aws-secret-access-key', is_secret=True)
    region = _config_get('aws-default-region', is_secret=True)
    cmd = f"AWS_DEFAULT_REGION={region} aws {cmd}"
    if len(access) >= 20 and len(secret) >= 40:
        cmd = f"AWS_ACCESS_KEY_ID={access} AWS_SECRET_ACCESS_KEY={secret} " + cmd
    return cmd

def aws_check_output(cmd):
    return subprocess.check_output(aws_process_cmd(cmd), shell=True)


def exec(cmd):
    subprocess.check_call(aws_process_cmd(cmd), shell=True)


def create_volume(disk_size_gb, labels, use_existing_disk_name=None, zone=0):
    assert not use_existing_disk_name, 'using existing disk name is not supported yet'
    availability_zone = get_storage_availability_zone(zone)
    logs.info(f'creating persistent disk with size {disk_size_gb} in availability zone {availability_zone}')
    tags = 'ResourceType=volume,Tags=[{Key=Owner,Value=cco}]'
    data = json.loads(aws_check_output(f'ec2 create-volume -- --size {disk_size_gb} --availability-zone {availability_zone} --tag-specification "{tags}"'))
    volume_id = data['VolumeId']
    logs.info(f'volume_id={volume_id}')
    kubectl.apply({
        'apiVersion': 'v1', 'kind': 'PersistentVolume',
        'metadata': {'name': volume_id, 'namespace': 'ckan-cloud'},
        'spec': {
            'storageClassName': '',
            'capacity': {'storage': f'{disk_size_gb}G'},
            'accessModes': ['ReadWriteOnce'],
            'awsElasticBlockStore': {'volumeID': volume_id}
        }
    })
    kubectl.apply({
        'apiVersion': 'v1', 'kind': 'PersistentVolumeClaim',
        'metadata': {'name': volume_id, 'namespace': 'ckan-cloud'},
        'spec': {
            'storageClassName': '',
            'volumeName': volume_id,
            'accessModes': ['ReadWriteOnce'],
            'resources': {'requests': {'storage': f'{disk_size_gb}G'}}
        }
    })
    return {
        'persistentVolumeClaim': {
            'claimName': volume_id
        },
        'nodeSelector': {
            'failure-domain.beta.kubernetes.io/zone': availability_zone
        }
    }


def get_storage_availability_zone(zone):
    region = _config_get('aws-default-region', is_secret=True)
    region += 'abc'[zone % 3]
    return region


def get_name():
    name = _config_get('cluster-name')
    if not name:
        name = get_info()['name']
        _config_set('cluster-name', name)
    return name


def get_boto3_client(service_name):
    import boto3
    access = _config_get('aws-access-key-id', is_secret=True)
    secret = _config_get('aws-secret-access-key', is_secret=True)
    kwargs = {}
    if len(access) >= 20 and len(secret) >= 40:
        kwargs = dict(aws_access_key_id=access, aws_secret_access_key=secret)
    return boto3.client(service_name, **kwargs)


def get_dns_hosted_zone_id(root_domain):
    client = get_boto3_client('route53')
    return client.list_hosted_zones_by_name(DNSName=f'{root_domain}.', MaxItems='1')['HostedZones'][0]['Id']


def update_dns_record(sub_domain, root_domain, load_balancer_hostname):
    logs.info('updating Route53 DNS record', sub_domain=sub_domain, root_domain=root_domain, load_balancer_hostname=load_balancer_hostname)
    hosted_zone_id = get_dns_hosted_zone_id(root_domain)
    logs.info(hosted_zone_id=hosted_zone_id)
    response = get_boto3_client('route53').change_resource_record_sets(
        HostedZoneId=hosted_zone_id,
        ChangeBatch={
            'Comment': 'ckan-cloud-operator',
            'Changes': [
                {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'Name': f'{sub_domain}.{root_domain}.',
                        'Type': 'CNAME',
                        'TTL': 300,
                        'ResourceRecords': [
                            {
                                'Value': load_balancer_hostname
                            },
                        ],
                    }
                },
            ]
        }
    )
    assert response['ResponseMetadata']['HTTPStatusCode'] == 200



import yaml
import subprocess
import tempfile
import traceback
import os
import time


def activate_auth(project, compute_zone, service_account_email, service_account_json):
    """Authenticate with gcloud"""
    args = [project, compute_zone, service_account_email, service_account_json]
    assert all(args), f'missing gcloud auth details: {args}'
    with tempfile.NamedTemporaryFile(suffix='.json', delete=False) as f:
        f.write(service_account_json.encode())
    try:
        check_call(
            project, compute_zone,
            f'auth activate-service-account {service_account_email} --key-file={f.name}'
            ' && '
            f'gcloud --project={project} config set compute/zone {compute_zone}'
        )
    except Exception:
        traceback.print_exc()
    os.unlink(f.name)


def check_output(project, compute_zone, cmd, gsutil=False):
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.check_output(f'{bin} {cmd}', shell=True)


def run(project, compute_zone, cmd, gsutil=False, **kwargs):
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.run(f'{bin} {cmd}', shell=True, **kwargs)


def call(project, compute_zone, cmd, gsutil=False, **kwargs):
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.call(f'{bin} {cmd}', shell=True, **kwargs)


def check_call(project, compute_zone, cmd, gsutil=False):
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.check_call(f'{bin} {cmd}', shell=True)


def getstatusoutput(project, compute_zone, cmd, gsutil=False):
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.getstatusoutput(f'{bin} {cmd}')


def _set_gcloud_storage_sql_permissions(project, compute_zone, sql_instance_name, import_url):
    print('setting permissions to cloud storage for import to sql')
    gcloud_sql_instance = yaml.load(check_output(
        project, compute_zone,
        f'sql instances describe {sql_instance_name}',
    ))
    gcloud_sql_service_account_email = gcloud_sql_instance['serviceAccountEmailAddress']
    check_call(
        project, compute_zone,
        f'acl ch -u {gcloud_sql_service_account_email}:R {import_url}',
        gsutil=True
    )

def _import_gcloud_sql_db(project, compute_zone, sql_instance_name, import_url, db_name, import_user):
    print(f'Importing Gcloud SQL from: {import_url}')
    _set_gcloud_storage_sql_permissions(project, compute_zone, sql_instance_name, import_url)
    while True:
        proc = run(
            project, compute_zone,
            f'--quiet sql import sql --async {sql_instance_name} {import_url} --database={db_name} --user={import_user} ',
            capture_output=True
        )
        error = proc.stderr.decode()
        print(error)
        if 'HTTPError 409:' in error:
            print('Waiting for other operation to complete...')
            time.sleep(60)
            continue
        break
    output = proc.stdout.decode()
    if proc.returncode == 0:
        print(output)
        operation_id = output.strip().split('/')[-1]
        print(f'Waiting for sql import operation {operation_id} to complete...')
        while True:
            time.sleep(5)
            operation = None
            try:
                operation = yaml.load(check_output(project, compute_zone, f'sql operations describe {operation_id}'))
                print(operation['status'])
                if operation['status'] == 'DONE':
                    if operation.get('error'):
                        print(yaml.dump(operation['error'], default_flow_style=False))
                        raise Exception('sql import operation failed')
                    else:
                        print('operation completed successfully')
                        print(operation)
                        break
            except Exception:
                print(operation)
                raise
    else:
        print(f'returncode={proc.returncode}')
        print(output)
        raise Exception(f'failed to import: {import_url} --> {db_name}   ({import_user})')



from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG, getLevelName
import datetime
from distutils.util import strtobool
import os
from ruamel import yaml
from ruamel.yaml.serializer import Serializer as ruamelSerializer
from ruamel.yaml.emitter import Emitter as ruamelEmitter
import sys
import subprocess

CKAN_CLOUD_OPERATOR_DEBUG = strtobool(os.environ.get('CKAN_CLOUD_OPERATOR_DEBUG', 'n'))
CKAN_CLOUD_OPERATOR_DEBUG_FILE = os.environ.get('CKAN_CLOUD_OPERATOR_DEBUG_FILE', '').strip()
CKAN_CLOUD_OPERATOR_DEBUG_VERBOSE = strtobool(os.environ.get('CKAN_CLOUD_OPERATOR_DEBUG_VERBOSE', 'n'))

DEBUG_VERBOSE = 'verbose debug'


def info(*args, **kwargs):
    log(INFO, *args, **kwargs)


def debug(*args, **kwargs):
    log(DEBUG, *args, **kwargs)


def debug_verbose(*args, **kwargs):
    log(DEBUG_VERBOSE, yaml.dump([args, kwargs], default_flow_style=False))


def warning(*args, **kwargs):
    log(WARNING, *args, **kwargs)


def error(*args, **kwargs):
    log(ERROR, *args, **kwargs)


def critical(*args, **kwargs):
    log(CRITICAL, *args, **kwargs)


def log(level, *args, **kwargs):
    if not _skip_log_level(level):
        _print_log_msg(level, _get_log_msg(level, *args, **kwargs))


def important_log(level, *args, **kwargs):
    if not _skip_log_level(level):
        _print_log_msg(level, _get_important_log_msg(level, *args, **kwargs))


def exit_great_success(quiet=False):
    if not quiet:
        info('Great Success!')
    exit(0)


def exit_catastrophic_failure(exitcode=1, quiet=False):
    if not quiet:
        critical('Catastrophic Failure!')
    exit(exitcode)

# subprocess

def log_subprocess_output(stdout, stderr):
    for line in stderr.decode('utf8').split('\n'):
        if line:
            warning(line)
    for line in stdout.decode('utf8').split('\n'):
        if line:
            info(line)


def subprocess_run(command, input=None):
    completed = subprocess.run(
        command, input=input, 
        shell=True, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    log_subprocess_output(completed.stdout, completed.stderr)
    completed.check_returncode()

def subprocess_check_output(*args, **kw):
    try:
        return subprocess.check_output(*args, stderr=subprocess.PIPE, **kw)
    except subprocess.CalledProcessError as e:
        log_subprocess_output(e.stdout, e.stderr)
        raise

def subprocess_check_call(*args, **kw):
    try:
        return subprocess.check_call(*args, stderr=subprocess.PIPE, stdout=subprocess.PIPE, **kw)
    except subprocess.CalledProcessError as e:
        log_subprocess_output(e.stdout, e.stderr)
        raise

# yaml dumping


def debug_yaml_dump(*args, **kwargs):
    if len(args) == 1:
        debug(yaml.dump(args[0], Dumper=YamlSafeDumper, default_flow_style=False), **kwargs)
    else:
        debug(yaml.dump(args, Dumper=YamlSafeDumper, default_flow_style=False), **kwargs)


def print_yaml_dump(data, exit_success=False):
    yaml.dump(data, sys.stdout, Dumper=YamlSafeDumper, default_flow_style=False)
    if exit_success:
        exit_great_success(quiet=True)


def yaml_dump(data, *args, **kwargs):
    return yaml.dump(data, *args, Dumper=YamlSafeDumper, default_flow_style=False, **kwargs)


class YamlSafeDumper(yaml.SafeDumper):

    def ignore_aliases(self, data):
        return True

    def represent_undefined(self, data):
        return None


# private functions


def _get_log_msg(level, *args, **kwargs):
    msg = datetime.datetime.now().strftime('%Y-%m-%d %H:%M') + ' ' + _get_level_name(level) + ' '
    if len(kwargs) > 0:
        msg += '(' + ','.join([f'{k}="{v}"' for k, v in kwargs.items()]) + ') '
    msg += ' '.join(args)
    return msg


def _get_important_log_msg(level, *args, **kwargs):
    msg = ''
    header = datetime.datetime.now().strftime('%Y-%m-%d %H:%M') + ' ' + _get_level_name(level)
    msg += f'\n{header}\n'
    if len(kwargs) > 0:
        metadata = '(' + ','.join([f'{k}="{v}"' for k, v in kwargs.items()]) + ')'
        msg += metadata + '\n'
    msg += '\n\n'
    if len(args) > 0:
        title = args[0]
        msg += f'== {title}\n'
        if len(args) > 1:
            msg += ' '.join(args[1:])
            msg += '\n'
    return msg


def _get_level_name(level):
    if level == DEBUG_VERBOSE:
        return getLevelName(DEBUG)
    else:
        return getLevelName(level)


def _skip_log_level(level):
    return (
           (level == DEBUG and not CKAN_CLOUD_OPERATOR_DEBUG and not CKAN_CLOUD_OPERATOR_DEBUG_FILE)
        or (level == DEBUG_VERBOSE and not CKAN_CLOUD_OPERATOR_DEBUG_VERBOSE and not CKAN_CLOUD_OPERATOR_DEBUG_FILE)
    )


def _print_log_msg(level, msg):
    if CKAN_CLOUD_OPERATOR_DEBUG_FILE:
        with open(CKAN_CLOUD_OPERATOR_DEBUG_FILE, 'a') as f:
            print(msg, file=f)
    if (
        (level == DEBUG and (CKAN_CLOUD_OPERATOR_DEBUG or CKAN_CLOUD_OPERATOR_DEBUG_VERBOSE))
        or (level == DEBUG_VERBOSE and CKAN_CLOUD_OPERATOR_DEBUG_VERBOSE)
        or level not in [DEBUG, DEBUG_VERBOSE]
    ):
        print(msg, file=sys.stderr)
        sys.stderr.flush()



import yaml
import time
import datetime
import binascii
import os
import traceback

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.labels import manager as labels_manager
from .deployment import manager as deployment_manager
from ckan_cloud_operator.routers import manager as routers_manager

from .constants import APP_CRD_KIND_SUFFIX, APP_CRD_PLURAL_SUFFIX, APP_CRD_SINGULAR
from .constants import APP_NAME_CRD_KIND_SUFFIX, APP_NAME_CRD_PLURAL_SUFFIX, APP_NAME_CRD_SINGULAR


def initialize(interactive=False):
    crds_manager.install_crd(APP_CRD_SINGULAR, APP_CRD_PLURAL_SUFFIX, APP_CRD_KIND_SUFFIX)
    crds_manager.install_crd(APP_NAME_CRD_SINGULAR, APP_NAME_CRD_PLURAL_SUFFIX, APP_NAME_CRD_KIND_SUFFIX)


def create(deployment_provider, instance_id=None, instance_name=None, values=None, values_filename=None, exists_ok=False,
           dry_run=False, update_=False, wait_ready=False, skip_deployment=False, skip_route=False, force=False):
    assert deployment_provider in ['helm']
    if not instance_id:
        if instance_name:
            instance_id = '{}-{}'.format(instance_name, _generate_password(6))
            logs.info('Generated instance id based on instance name', instance_name=instance_name, instance_id=instance_id)
        else:
            instance_id = _generate_password(12)
            logs.info('Generated instance id', instance_id=instance_id)
    if values_filename:
        assert values is None
        with open(values_filename) as f:
            values = yaml.load(f.read())
    if not exists_ok and crds_manager.get(APP_CRD_SINGULAR, name=instance_id, required=False):
        raise Exception('instance already exists')
    values_id = values.get('id')
    if values_id and values_id != instance_id:
        logs.warning(f'changing instance id in spec from {values_id} to the instance id {instance_id}')
    values.update(id=instance_id)
    logs.info('Creating instance', instance_id=instance_id)
    instance = crds_manager.get_resource(
        APP_CRD_SINGULAR, instance_id,
        extra_label_suffixes={'deployment-provider': deployment_provider},
        spec=values
    )
    label_prefix = labels_manager.get_label_prefix()
    ckan_cloud_annotations = {
        f'{label_prefix}/deployment-provider': deployment_provider,
        f'{label_prefix}/instance-id': instance_id
    }
    logs.info('setting ckan-cloud annotations', ckan_cloud_annotations=ckan_cloud_annotations)
    instance['metadata'].setdefault('annotations', {}).update(**ckan_cloud_annotations)
    kubectl.apply(instance, dry_run=dry_run)
    if instance_name:
        set_name(instance_id, instance_name, dry_run=dry_run)
    if update_:
        update(instance_id, wait_ready=wait_ready, skip_deployment=skip_deployment, skip_route=skip_route, force=force,
               dry_run=dry_run)
    return instance_id


def update(instance_id_or_name, override_spec=None, persist_overrides=False, wait_ready=False, skip_deployment=False,
           skip_route=False, force=False, dry_run=False):
    instance_id, instance = _get_instance(instance_id_or_name, required=not dry_run)
    if dry_run:
        logs.info('update instance', instance_id=instance_id, instance_id_or_name=instance_id_or_name,
                  override_spec=override_spec, persist_overrides=persist_overrides, wait_ready=wait_ready,
                  skip_deployment=skip_deployment, skip_route=skip_route, force=force, dry_run=dry_run)
    else:
        pre_update_hook_data = deployment_manager.pre_update_hook(instance_id, instance, override_spec,
                                                                  skip_route)
        if persist_overrides:
            logs.info('Persisting overrides')
            kubectl.apply(instance)
        if not skip_deployment:
            deployment_manager.update(instance_id, instance)
            if wait_ready:
                wait_instance_ready(instance_id_or_name)
        if not skip_route and pre_update_hook_data.get('sub-domain'):
            root_domain = pre_update_hook_data.get('root-domain')
            sub_domain = pre_update_hook_data['sub-domain']
            assert root_domain == routers_manager.get_default_root_domain(), \
                'invalid domain, must use default root domain'
            logs.info(f'adding instance default route to {sub_domain}.{root_domain}')
            routers_manager.create_subdomain_route('instances-default', {
                'target-type': 'app-instance',
                'app-instance-id': instance_id,
                'root-domain': root_domain,
                'sub-domain': sub_domain
            })
            routers_manager.update('instances-default', wait_ready)
        else:
            logs.info('skipping route creation', skip_route=skip_route,
                      sub_domain=pre_update_hook_data.get('sub-domain'))
        logs.info('Instance is ready', instance_id=instance_id,
                  instance_name=(instance_id_or_name if instance_id_or_name != instance_id else None))


def delete(instance_id):
    try:
        instance_id, instance = _get_instance(instance_id=instance_id)
    except Exception:
        logs.error(traceback.format_exc())
        instance, instance_type = None, None
    try:
        deployment_manager.delete(instance_id, instance)
    except Exception as e:
        logs.error('error during deployment delete', error=str(e))
    try:
        crds_manager.delete(APP_CRD_SINGULAR, instance_id)
    except Exception as e:
        logs.error('error during crd delete', error=str(e))


def delete_instances(instance_ids_or_names=None, dry_run=True, instance_ids=None):
    if instance_ids:
        assert not instance_ids_or_names
        use_instance_ids = True
    else:
        assert not instance_ids
        use_instance_ids = False
    logs.info(use_instance_ids=use_instance_ids, instance_ids=instance_ids, instance_ids_or_names=instance_ids_or_names)
    for instance_id_or_name in (instance_ids if use_instance_ids else instance_ids_or_names):
        logs.info(instance_id_or_name=instance_id_or_name)
        if use_instance_ids:
            instance_id = instance_id_or_name
        else:
            instance_id, _ = _get_instance(instance_id_or_name, required=False)
        instance_name = instance_id_or_name if instance_id != instance_id_or_name else None
        yield {
            'id': instance_id,
            **({'name': instance_name} if instance_name else {})
        }
        errors = []
        if instance_id:
            if not dry_run:
                try:
                    delete(instance_id=instance_id)
                except Exception as e:
                    errors.append(str(e))
                    logs.error('exception raised during instance deletion, '
                               'this does not always indicate deletion was not successful',
                               exception=str(e))
        else:
            errors.append(f'Failed to get instance_id for instance_id_or_name {instance_id_or_name}')
        if instance_name:
            try:
                delete_name(instance_name)
            except Exception as e:
                errors.append(str(e))
                logs.error('exception raised during instance name deletion.', exception=str(e))
        yield {'errors': errors}


def wait_instance_ready(instance_id_or_name):
    logs.info(f'Waiting for instance ready status ({instance_id_or_name})')
    time.sleep(3)
    while True:
        data = get(instance_id_or_name)
        if data.get('ready'):
            break
        else:
            logs.print_yaml_dump(
                {
                    k: v for k, v in data.items()
                    if (k not in ['ready'] and type(v) == dict and not v.get('ready')) or k == 'namespace'
                }
            )
            time.sleep(2)


def edit(instance_id_or_name):
    instance_id, instance = _get_instance(instance_id_or_name)
    crds_manager.edit(APP_CRD_SINGULAR, name=instance_id)


def get(instance_id_or_name, attr=None, exclude_attr=None, with_spec=False):
    """Get detailed information about the instance and related components"""
    instance_id, instance = _get_instance(instance_id_or_name)
    if not exclude_attr:
        exclude_attr = []
    if not with_spec:
        exclude_attr.append('spec')
    gets = {
        'deployment': lambda: deployment_manager.get(instance_id, instance),
        'spec': lambda: instance['spec'],
        'routes': lambda: {
            'domain': instance['spec'].get('domain'),
            'backend-url': get_backend_url(instance_id=instance_id),
            'routes': [
                {
                    'sub-domain': r['spec'].get('sub-domain'),
                    'root-domain': r['spec'].get('root-domain'),
                    'router_name': r['spec'].get('routers_name'),
                    'name': r['spec'].get('name'),
                }
                for r
                in routers_manager.get_app_instance_routes(instance_id)],
            'ready': True
        }
    }
    if exclude_attr:
        gets = {k: v for k, v in gets.items() if k not in exclude_attr}
    if attr:
        return gets[attr]()
    else:
        ret = {'ready': True}
        for k, v in gets.items():
            try:
                ret[k] = v()
            except Exception as e:
                ret[k] = {'ready': False, 'errors': [str(e)]}
            if type(ret[k]) == dict and not ret[k].get('ready'):
                ret['ready'] = False
        ret['id'] = instance_id
        return ret


def get_backend_url(instance_id_or_name=None, instance_id=None):
    return deployment_manager.get_backend_url(*_get_instance(instance_id_or_name, instance_id=instance_id))


def get_all_instance_id_names():
    instance_names = crds_manager.get(APP_NAME_CRD_SINGULAR, required=False)
    instance_name_ids = {}
    if instance_names:
        for instance_name in instance_names['items']:
            instance_name_ids[instance_name['spec']['latest-instance-id']] = instance_name['spec']['name']
    label_prefix = labels_manager.get_label_prefix()
    for instance in crds_manager.get(APP_CRD_SINGULAR)['items']:
        instance_id = instance['metadata']['labels'][f'{label_prefix}/crd-app-name']
        instance_name = instance_name_ids.pop(instance_id, None)
        yield {'id': instance_id, 'name': instance_name}
    for instance_id, instance_name in instance_name_ids.items():
        yield {'id': instance_id, 'name': instance_name}


def list_instances(full=False, quick=False, name=None):
    for instance in get_all_instance_id_names():
        if name is not None and instance['name'] != name:
            continue
        if quick:
            yield {**instance, 'ready': None}
        else:
            try:
                instance_data = get(instance['id'])
            except Exception:
                instance_data = {}
            if full:
                instance_data['name'] = instance['name']
                yield instance_data
            else:
                yield {**instance, 'ready': instance_data.get('ready')}


def delete_name(instance_name):
    crds_manager.delete(APP_NAME_CRD_SINGULAR, instance_name)


def set_name(instance_id, instance_name, dry_run=False):
    resource = crds_manager.get(APP_NAME_CRD_SINGULAR, name=instance_name, required=False)
    if resource:
        resource['spec']['latest-instance-id'] = instance_id
        if not resource['spec']['instance-ids'].get(instance_id):
            resource['spec']['instance-ids'][instance_id] = {'added': datetime.datetime.now()}
    else:
        resource = crds_manager.get_resource(
            APP_NAME_CRD_SINGULAR, instance_name,
            spec={
                'name': instance_name,
                'latest-instance-id': instance_id,
                'instance-ids': {
                    instance_id: {'added': datetime.datetime.now()}
                }
            }
        )
    if dry_run:
        logs.print_yaml_dump(resource)
    else:
        kubectl.apply(resource)


def _get_instance(instance_id_or_name=None, instance_id=None, required=True):
    if instance_id:
        logs.debug(f'Getting instance using instance_id', instance_id=instance_id)
        instance = crds_manager.get(APP_CRD_SINGULAR, name=instance_id, required=False)
        instance_name = None
    else:
        logs.debug(f'Attempting to get instance using id', instance_id_or_name=instance_id_or_name)
        instance = crds_manager.get(APP_CRD_SINGULAR, name=instance_id_or_name, required=False)
        if instance:
            instance_id = instance_id_or_name
            instance_name = None
        else:
            logs.debug(f'Attempting to get instance from instance name', instance_id_or_name=instance_id_or_name)
            instance_name = crds_manager.get(APP_NAME_CRD_SINGULAR, name=instance_id_or_name, required=False)
            if instance_name:
                instance_id = instance_name['spec'].get('latest-instance-id')
                logs.debug(instance_id=instance_id)
                instance = crds_manager.get(APP_CRD_SINGULAR, name=instance_id, required=False)
                instance_name = instance_id_or_name
            else:
                instance_name = None
    logs.debug_yaml_dump(instance_name=instance_name, instance_id=instance_id, instance=bool(instance))
    if required:
        assert instance_id and len(instance) > 2, \
            f'Failed to find instance (instance_id_or_name={instance_id_or_name}, instance_id={instance_id})'
    return instance_id, instance


def _generate_password(length):
    return binascii.hexlify(os.urandom(length)).decode()



from ckan_cloud_operator import logs
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.labels import manager as labels_manager
from ckan_cloud_operator.annotations import manager as annotations_manager


def get_provider(submodule, required=True, supported_provider_ids=None, default=None, verbose=False, provider_id=None):
    if default: required = False
    if not provider_id:
        provider_id = get_provider_id(submodule, required=required)
        if not provider_id:
            if default:
                provider_id = default
            else:
                return None
        if verbose:
            logs.info(f'submodule {submodule} using provider_id {provider_id}')
    if not supported_provider_ids or provider_id in supported_provider_ids:
        provider_manager = _get_submodule_ids_provider_or_provider_ids(submodule, provider_id)
    else:
        provider_manager = None
    if provider_manager:
        return provider_manager
    else:
        msg = f' (supported provider ids: {supported_provider_ids})' if supported_provider_ids else ''
        logs.critical(f'Invalid submodule / provider: {submodule} / {provider_id}{msg}')
        raise Exception('failed to get provider')


def get_providers(submodule, verbose=False):
    res = {
        provider_id: _get_submodule_ids_provider_or_provider_ids(submodule, provider_id)
        for provider_id in _get_submodule_ids_provider_or_provider_ids(submodule)
    }
    if verbose:
        provider_ids = list(res.keys())
        logs.info(f'Providers for submodule {submodule}: {provider_ids}')
    return res


def list_providers():
    res = {}
    for submodule in _get_submodule_ids_provider_or_provider_ids():
        res[submodule] = {
            'providers': _get_submodule_ids_provider_or_provider_ids(submodule),
            'selected-provider': get_provider_id(submodule, required=False)
        }
    return res


def set_provider(submodule, provider_id):
    config_manager.set(
        key=get_operator_configmap_key(submodule, suffix='main-provider-id'),
        value=provider_id
    )


def get_provider_id(submodule, required=True, default=None):
    if default: required = False
    return config_manager.get(
        get_operator_configmap_key(submodule, suffix='main-provider-id'),
        required=required,
        default=default
    )


def get_operator_configmap_key(submodule, provider_id=None, suffix=None):
    """Returns a key which can be used to store data for the given submodule/provider in the default operator configmap"""
    return get_resource_name(submodule, provider_id, suffix)


def config_set(submodule, provider_id=None, key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None):
    """store key/values in a secret or configmap"""
    resource_name = get_resource_name(submodule, provider_id=provider_id, suffix=suffix)
    config_manager.set(
        key=key,
        value=value,
        values=values,
        secret_name=resource_name if is_secret else None,
        configmap_name=None if is_secret else resource_name,
        namespace=namespace
    )


def config_get(submodule, provider_id=None, key=None, default=None, required=False, namespace=None, is_secret=False,
               suffix=None, template=None):
    resource_name = get_resource_name(submodule, provider_id=provider_id, suffix=suffix)
    return config_manager.get(
        key=key,
        default=default,
        secret_name=resource_name if is_secret else None,
        configmap_name=None if is_secret else resource_name,
        required=required,
        namespace=namespace,
        template=template
    )


def config_get_volume_spec(submodule, provider_id, volume_name, is_secret=False, suffix=None):
    resource_name = get_resource_name(submodule, provider_id=provider_id, suffix=suffix)
    if is_secret:
        return {'name': volume_name, 'secret': {'secretName': resource_name}}
    else:
        return {'name': volume_name, 'configMap': {'name': resource_name}}


def config_interactive_set(submodule, provider_id=None, default_values=None, namespace=None,
                           is_secret=False, suffix=None, from_file=False, interactive=True):
    resource_name = get_resource_name(submodule, provider_id=provider_id, suffix=suffix)
    config_manager.interactive_set(
        default_values,
        secret_name=resource_name if is_secret else None,
        configmap_name=None if is_secret else resource_name,
        namespace=namespace,
        from_file=from_file,
        interactive=interactive
    )


def get_resource_name(submodule, provider_id=None, suffix=None, short=False):
    return labels_manager.get_resource_name(get_resource_suffix(submodule, provider_id=provider_id, suffix=suffix,
                                                                short=short), short=short)


def get_resource_suffix(submodule, provider_id=None, suffix=None, short=False):
    if short:
        res = _get_submodule_provider_config(submodule, provider_id).get('short-resource-suffix')
        assert res, f'Failed to get short-resource-suffix ({submodule}={provider_id})'
        if suffix:
            res += suffix
        return res
    else:
        parts = ['provider', submodule]
        if provider_id:
            parts.append(provider_id)
        if suffix:
            parts.append(suffix)
        return '-'.join(parts)


def get_resource_labels(submodule, provider_id, extra_label_suffixes=None, for_deployment=False, suffix=None):
    label_suffixes = {
        'provider-submodule': submodule,
        'provider-id': provider_id,
    }
    if suffix:
        label_suffixes['provider-submodule-suffix'] = suffix
    if extra_label_suffixes:
        label_suffixes.update(**extra_label_suffixes)
    extra_labels = {'app': get_deployment_app_label(submodule, provider_id, suffix=suffix)} if for_deployment else {}
    return labels_manager.get_resource_labels(label_suffixes, extra_labels=extra_labels)


def get_deployment_app_label(submodule, provider_id, suffix=None):
    return get_resource_suffix(submodule, provider_id=provider_id, suffix=suffix)


def get_resource_annotations(submodule, provider_id=None, suffix=None, with_timestamp=True):
    return annotations_manager.get_global_annotations(with_timestamp=with_timestamp)


def _get_submodule_provider_config(submodule, provider_id):
    return {
        'apps-deployment': {
            'helm': {
                'short-resource-suffix': 'app-helm'
            }
        }
    }.get(submodule, {}).get(provider_id, {})


def _get_submodule_ids_provider_or_provider_ids(submodule=None, provider_id=None):
    from ckan_cloud_operator.providers.db.proxy.constants import PROVIDER_SUBMODULE as db_proxy_provider_submodule
    from ckan_cloud_operator.providers.db.constants import PROVIDER_SUBMODULE as db_provider_submodule
    from ckan_cloud_operator.providers.db.web_ui.constants import PROVIDER_SUBMODULE as db_web_ui_submodule
    from ckan_cloud_operator.providers.users.constants import PROVIDER_SUBMODULE as users_provider_submodule
    from ckan_cloud_operator.providers.cluster.constants import PROVIDER_SUBMODULE as cluster_provider_submodule
    from ckan_cloud_operator.providers.storage.constants import PROVIDER_SUBMODULE as storage_provider_submodule
    from ckan_cloud_operator.providers.solr.constants import PROVIDER_SUBMODULE as solr_provider_submodule

    if not submodule:
        return [
            db_proxy_provider_submodule,
            db_provider_submodule,
            db_web_ui_submodule,
            users_provider_submodule,
            cluster_provider_submodule,
            storage_provider_submodule,
            solr_provider_submodule
        ]

    ## db-proxy

    elif submodule == db_proxy_provider_submodule:
        from ckan_cloud_operator.providers.db.proxy.pgbouncer.constants import PROVIDER_ID as pgbouncer_provider_id
        from ckan_cloud_operator.providers.db.proxy.gcloudsql.constants import PROVIDER_ID as db_proxy_gcloudsql_provider_id
        from ckan_cloud_operator.providers.db.proxy.rds.constants import PROVIDER_ID as db_proxy_rds_provider_id

        if not provider_id:
            return [pgbouncer_provider_id, db_proxy_gcloudsql_provider_id, db_proxy_rds_provider_id]

        ## pgbouncer

        elif provider_id == pgbouncer_provider_id:
            from ckan_cloud_operator.providers.db.proxy.pgbouncer import manager as pgbouncer_manager

            return pgbouncer_manager

        ## gcloudsql proxy

        elif provider_id == db_proxy_gcloudsql_provider_id:
            from ckan_cloud_operator.providers.db.proxy.gcloudsql import manager as db_proxy_gcloudsql_manager

            return db_proxy_gcloudsql_manager

        ## rds proxy

        elif provider_id == db_proxy_rds_provider_id:
            from ckan_cloud_operator.providers.db.proxy.rds import manager as db_proxy_rds_manager

            return db_proxy_rds_manager

    ## db

    elif submodule == db_provider_submodule:
        from ckan_cloud_operator.providers.db.azuresql.constants import PROVIDER_ID as azuresql_provider_id
        from ckan_cloud_operator.providers.db.gcloudsql.constants import PROVIDER_ID as gcloudsql_provider_id
        from ckan_cloud_operator.providers.db.rds.constants import PROVIDER_ID as rds_provider_id
        from ckan_cloud_operator.providers.db.minikube.constants import PROVIDER_ID as minikube_provider_id

        if not provider_id:
            return [gcloudsql_provider_id]

        ## gcloud

        elif provider_id == gcloudsql_provider_id:
            from ckan_cloud_operator.providers.db.gcloudsql import manager as gcloudsql_manager

            return gcloudsql_manager

        ## rds

        elif provider_id == rds_provider_id:
            from ckan_cloud_operator.providers.db.rds import manager as rds_manager

            return rds_manager

        ## azuresql

        elif provider_id == azuresql_provider_id:
            from ckan_cloud_operator.providers.db.azuresql import manager as azuresql_manager

            return azuresql_manager

        ## minikube

        elif provider_id == minikube_provider_id:
            from ckan_cloud_operator.providers.db.minikube import manager as minikube_manager

            return minikube_manager

    ## db-web-ui

    elif submodule == db_web_ui_submodule:
        from ckan_cloud_operator.providers.db.web_ui.adminer.constants import PROVIDER_ID as adminer_provider_id

        if not provider_id:
            return [adminer_provider_id]

        ## adminer

        elif provider_id == adminer_provider_id:
            from ckan_cloud_operator.providers.db.web_ui.adminer import manager as adminer_manager

            return adminer_manager

    ## users

    elif submodule == users_provider_submodule:
        from ckan_cloud_operator.providers.users.gcloud.constants import PROVIDER_ID as users_gcloud_provider_id
        from ckan_cloud_operator.providers.users.rancher.constants import PROVIDER_ID as users_rancher_provider_id

        if not provider_id:
            return [users_gcloud_provider_id, users_rancher_provider_id]

        ## gcloud

        elif provider_id == users_gcloud_provider_id:
            from ckan_cloud_operator.providers.users.gcloud import manager as users_gcloud_manager

            return users_gcloud_manager

        ## rancher

        elif provider_id == users_rancher_provider_id:
            from ckan_cloud_operator.providers.users.rancher import manager as users_rancher_manager

            return users_rancher_manager

    ## cluster

    elif submodule == cluster_provider_submodule:
        from ckan_cloud_operator.providers.cluster.gcloud.constants import PROVIDER_ID as cluster_gcloud_provider_id
        from ckan_cloud_operator.providers.cluster.aws.constants import PROVIDER_ID as cluster_aws_provider_id
        from ckan_cloud_operator.providers.cluster.azure.constants import PROVIDER_ID as cluster_azure_provider_id
        from ckan_cloud_operator.providers.cluster.minikube.constants import PROVIDER_ID as cluster_minikube_provider_id

        if not provider_id:
            return [cluster_gcloud_provider_id, cluster_aws_provider_id, cluster_azure_provider_id]

        ## gcloud

        elif provider_id == cluster_gcloud_provider_id:
            from ckan_cloud_operator.providers.cluster.gcloud import manager as clouster_gcloud_manager

            return clouster_gcloud_manager

        ## aws

        elif provider_id == cluster_aws_provider_id:
            from ckan_cloud_operator.providers.cluster.aws import manager as cluster_aws_manager

            return cluster_aws_manager

        ## azure

        elif provider_id == cluster_azure_provider_id:
            from ckan_cloud_operator.providers.cluster.azure import manager as cluster_azure_manager

            return cluster_azure_manager

        ## minikube

        elif provider_id == cluster_minikube_provider_id:
            from ckan_cloud_operator.providers.cluster.minikube import manager as cluster_minikube_manager

            return cluster_minikube_manager

    ## storage

    elif submodule == storage_provider_submodule:
        from ckan_cloud_operator.providers.storage.azure.constants import PROVIDER_ID as storage_azure_provider_id
        from ckan_cloud_operator.providers.storage.gcloud.constants import PROVIDER_ID as storage_gcloud_provider_id
        from ckan_cloud_operator.providers.storage.minio.constants import PROVIDER_ID as storage_minio_provider_id
        from ckan_cloud_operator.providers.storage.s3.constants import PROVIDER_ID as storage_s3_provider_id

        if not provider_id:
            return []

        # Azure
        elif provider_id == storage_azure_provider_id:
            from ckan_cloud_operator.providers.storage.azure import manager as storage_azure_manager

            return storage_azure_manager

        # Gcloud
        elif provider_id == storage_gcloud_provider_id:
            from ckan_cloud_operator.providers.storage.gcloud import manager as storage_gcloud_manager

            return storage_gcloud_manager

        # Minio
        elif provider_id == storage_minio_provider_id:
            from ckan_cloud_operator.providers.storage.minio import manager as storage_minio_manager

            return storage_minio_manager

        # AWS
        elif provider_id == storage_s3_provider_id:
            from ckan_cloud_operator.providers.storage.s3 import manager as storage_s3_manager

            return storage_s3_manager


    ## solr

    elif submodule == solr_provider_submodule:
        from ckan_cloud_operator.providers.solr.solrcloud.constants import PROVIDER_ID as solr_solrcloud_provider_id

        if not provider_id:
            return [solr_solrcloud_provider_id]

        ## solrcloud

        elif provider_id == solr_solrcloud_provider_id:
            from ckan_cloud_operator.providers.solr.solrcloud import manager as solr_solrcloud_manager

            return solr_solrcloud_manager

    if not provider_id:
        return []
    else:
        return None
